from django.db import migrations, models

import mayan.apps.storage.classes
import mayan.apps.storage.models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='StorageSharedUploadedFile',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True,
                        serialize=False, verbose_name='ID'
                    )
                ),
                (
                    'file', models.FileField(
                        storage=mayan.apps.storage.classes.DefinedStorageLazy(
                            name='storage__shareduploadedfile'
                        ), upload_to=mayan.apps.storage.models.upload_to,
                        verbose_name='File'
                    )
                ),
                (
                    'filename', models.CharField(
                        max_length=255, verbose_name='Filename'
                    )
                ),
                (
                    'datetime', models.DateTimeField(
                        auto_now_add=True, verbose_name='Date time'
                    )
                ),
            ],
            options={
                'verbose_name': 'Shared uploaded file',
                'verbose_name_plural': 'Shared uploaded files',
            },
        ),
    ]



from mayan.apps.appearance.classes import Icon

icon_duplicated_document_list = Icon(
    driver_name='fontawesome', symbol='clone'
)
icon_duplicated_document_scan = Icon(
    driver_name='fontawesome-dual-classes', primary_class='fas fa-clone',
    secondary_class='far fa-eye'
)



import hashlib
import logging

from furl import furl
from graphviz import Digraph

from django.apps import apps
from django.core import serializers
from django.db import IntegrityError, models
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _

from mayan.apps.acls.models import AccessControlList
from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.common.validators import validate_internal_name
from mayan.apps.documents.models import Document, DocumentType
from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.events.classes import EventManagerSave
from mayan.apps.events.decorators import method_event
from mayan.apps.file_caching.models import CachePartitionFile
from ..events import (
    event_workflow_template_created, event_workflow_template_edited
)
from ..literals import (
    STORAGE_NAME_WORKFLOW_CACHE, SYMBOL_MATH_CONDITIONAL,
    WORKFLOW_ACTION_ON_ENTRY
)
from ..managers import WorkflowManager

__all__ = ('Workflow', 'WorkflowRuntimeProxy')
logger = logging.getLogger(name=__name__)


class Workflow(ExtraDataModelMixin, models.Model):
    """
    Fields:
    * label - Identifier. A name/label to call the workflow
    """
    auto_launch = models.BooleanField(
        default=True, help_text=_(
            'Launch workflow when document is created.'
        ), verbose_name=_('Auto launch')
    )
    internal_name = models.CharField(
        db_index=True, help_text=_(
            'This value will be used by other apps to reference this '
            'workflow. Can only contain letters, numbers, and underscores.'
        ), max_length=255, unique=True, validators=[validate_internal_name],
        verbose_name=_('Internal name')
    )
    label = models.CharField(
        help_text=_('A short text to describe the workflow.'),
        max_length=255, unique=True, verbose_name=_('Label')
    )
    document_types = models.ManyToManyField(
        related_name='workflows', to=DocumentType, verbose_name=_(
            'Document types'
        )
    )

    objects = WorkflowManager()

    class Meta:
        ordering = ('label',)
        verbose_name = _('Workflow')
        verbose_name_plural = _('Workflows')

    def __str__(self):
        return self.label

    def get_absolute_url(self):
        return reverse(viewname='document_states:workflow_template_list')

    @cached_property
    def cache(self):
        Cache = apps.get_model(app_label='file_caching', model_name='Cache')
        return Cache.objects.get(
            defined_storage_name=STORAGE_NAME_WORKFLOW_CACHE
        )

    @cached_property
    def cache_partition(self):
        partition, created = self.cache.partitions.get_or_create(
            name='{}'.format(self.pk)
        )
        return partition

    def delete(self, *args, **kwargs):
        self.cache_partition.delete()
        return super().delete(*args, **kwargs)

    def document_types_add(self, queryset, _event_actor=None):
        for document_type in queryset.all():
            self.document_types.add(document_type)
            event_workflow_template_edited.commit(
                action_object=document_type,
                actor=_event_actor or self._event_actor,
                target=self
            )

    def document_types_remove(self, queryset, _event_actor=None):
        for document_type in queryset.all():
            self.document_types.remove(document_type)
            event_workflow_template_edited.commit(
                action_object=document_type,
                actor=_event_actor or self._event_actor,
                target=self
            )
            self.instances.filter(
                document__document_type_id=document_type.pk
            ).delete()

    def generate_image(
        self, maximum_layer_order=None, transformation_instance_list=None,
        user=None
    ):
        # `user` argument added for compatibility.
        cache_filename = '{}'.format(self.get_hash())

        try:
            self.cache_partition.get_file(filename=cache_filename)
        except CachePartitionFile.DoesNotExist:
            logger.debug(
                'workflow cache file "%s" not found', cache_filename
            )

            image = self.render()
            with self.cache_partition.create_file(filename=cache_filename) as file_object:
                file_object.write(image)
        else:
            logger.debug(
                'workflow cache file "%s" found', cache_filename
            )

        return cache_filename

    def get_api_image_url(self, *args, **kwargs):
        final_url = furl()
        final_url.args = kwargs
        final_url.path = reverse(
            viewname='rest_api:workflow-template-image',
            kwargs={'workflow_template_id': self.pk}
        )
        final_url.args['_hash'] = self.get_hash()

        return final_url.tostr()

    def get_document_types_not_in_workflow(self):
        return DocumentType.objects.exclude(pk__in=self.document_types.all())

    def get_hash(self):
        result = hashlib.sha256(
            serializers.serialize(format='json', queryset=(self,)).encode()
        )
        for state in self.states.all():
            result.update(state.get_hash().encode())

        for transition in self.transitions.all():
            result.update(transition.get_hash().encode())

        return result.hexdigest()

    def get_initial_state(self):
        try:
            return self.states.get(initial=True)
        except self.states.model.DoesNotExist:
            return None
    get_initial_state.short_description = _('Initial state')

    def launch_for(self, document):
        if document.document_type in self.document_types.all():
            try:
                logger.info(
                    'Launching workflow %s for document %s', self, document
                )
                workflow_instance = self.instances.create(document=document)
                initial_state = self.get_initial_state()
                if initial_state:
                    for action in initial_state.entry_actions.filter(enabled=True):
                        context = workflow_instance.get_context()
                        context.update(
                            {
                                'action': action
                            }
                        )
                        action.execute(
                            context=context,
                            workflow_instance=workflow_instance
                        )
            except IntegrityError:
                logger.info(
                    'Workflow %s already launched for document %s', self, document
                )
            else:
                logger.info(
                    'Workflow %s launched for document %s', self, document
                )
                return workflow_instance
        else:
            logger.error(
                'This workflow is not valid for the document type of the '
                'document.'
            )

    def render(self):
        diagram = Digraph(
            name='finite_state_machine', graph_attr={
                'rankdir': 'LR', 'splines': 'polyline'
            }, format='png'
        )

        action_cache = {}
        state_cache = {}
        transition_cache = []

        for state in self.states.all():
            state_cache['s{}'.format(state.pk)] = {
                'name': 's{}'.format(state.pk),
                'label': state.label,
                'initial': state.initial,
                'connections': {'origin': 0, 'destination': 0}
            }

            for action in state.actions.all():
                if action.has_condition():
                    action_label = '{} {}'.format(
                        SYMBOL_MATH_CONDITIONAL, action.label
                    )
                else:
                    action_label = action.label

                action_cache['a{}'.format(action.pk)] = {
                    'name': 'a{}'.format(action.pk),
                    'label': action_label,
                    'state': 's{}'.format(state.pk),
                    'when': action.when,
                }

        for transition in self.transitions.all():
            if transition.has_condition():
                transition_label = '{} {}'.format(
                    SYMBOL_MATH_CONDITIONAL, transition.label
                )
            else:
                transition_label = transition.label

            transition_cache.append(
                {
                    'tail_name': 's{}'.format(transition.origin_state.pk),
                    'head_name': 's{}'.format(transition.destination_state.pk),
                    'label': transition_label
                }
            )
            state_cache['s{}'.format(transition.origin_state.pk)]['connections']['origin'] = state_cache['s{}'.format(transition.origin_state.pk)]['connections']['origin'] + 1
            state_cache['s{}'.format(transition.destination_state.pk)]['connections']['destination'] += 1

        for key, value in state_cache.items():
            kwargs = {
                'name': value['name'],
                'label': value['label'],
                'shape': 'doublecircle' if value['connections']['origin'] == 0 or value['connections']['destination'] == 0 or value['initial'] else 'circle',
                'style': 'filled' if value['initial'] else '',
                'fillcolor': '#eeeeee',
            }
            diagram.node(**kwargs)

        for transition in transition_cache:
            diagram.edge(**transition)

        for key, value in action_cache.items():
            kwargs = {
                'name': value['name'],
                'label': value['label'],
                'shape': 'box',
            }
            diagram.node(**kwargs)
            diagram.edge(
                **{
                    'head_name': '{}'.format(value['name']),
                    'tail_name': '{}'.format(value['state']),
                    'label': 'On entry' if value['when'] == WORKFLOW_ACTION_ON_ENTRY else 'On exit',
                    'arrowhead': 'dot',
                    'dir': 'both',
                    'arrowtail': 'dot',
                    'style': 'dashed',
                }
            )

        return diagram.pipe()

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_workflow_template_created,
            'target': 'self',
        },
        edited={
            'event': event_workflow_template_edited,
            'target': 'self',
        }
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)


class WorkflowRuntimeProxy(Workflow):
    class Meta:
        proxy = True
        verbose_name = _('Workflow runtime proxy')
        verbose_name_plural = _('Workflow runtime proxies')

    def get_document_count(self, user):
        """
        Return the numeric count of documents executing this workflow.
        The count is filtered by access.
        """
        return AccessControlList.objects.restrict_queryset(
            permission=permission_document_view,
            queryset=Document.valid.filter(workflows__workflow=self),
            user=user
        ).count()



import os
from pathlib import Path

from django.conf import settings
from django.conf.urls import include, url
from django.template import loader
from django.template.base import Template
from django.template.context import Context
from django.utils.encoding import force_text
from django.utils.html import mark_safe
from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.serialization import yaml_dump, yaml_load
from mayan.apps.databases.classes import BaseBackend
from mayan.apps.task_manager.classes import Worker
from mayan.settings.literals import (
    DEFAULT_DATABASE_NAME, DEFAULT_DATABASE_PASSWORD, DEFAULT_DATABASE_USER,
    DEFAULT_DIRECTORY_INSTALLATION, DEFAULT_OS_USERNAME,
    DEFAULT_USER_SETTINGS_FOLDER, DOCKER_DIND_IMAGE_VERSION,
    DOCKER_LINUX_IMAGE_VERSION, DOCKER_MYSQL_IMAGE_VERSION,
    DOCKER_POSTGRES_IMAGE_VERSION, GUNICORN_LIMIT_REQUEST_LINE,
    GUNICORN_MAX_REQUESTS, GUNICORN_REQUESTS_JITTER, GUNICORN_TIMEOUT,
    GUNICORN_WORKER_CLASS, GUNICORN_WORKERS
)

from .settings import (
    setting_client_backend_arguments, setting_client_backend_enabled
)
from .utils import load_env_file


class ClientBackend(BaseBackend):
    _loader_module_name = 'client_backends'

    @classmethod
    def get_backend_instance(cls, name):
        backend_class = cls.get(name=name)
        kwargs = setting_client_backend_arguments.value.get(name, {})
        return backend_class(**kwargs)

    @classmethod
    def post_load_modules(cls):
        cls.register_url_patterns()
        cls.launch_backends()

    @classmethod
    def launch_backends(cls):
        for backend_name in setting_client_backend_enabled.value:
            cls.get_backend_instance(name=backend_name).launch()

    @classmethod
    def register_url_patterns(cls):
        # Hidden import.
        from .urls import urlpatterns

        for backend_name in setting_client_backend_enabled.value:
            backend_instance = cls.get_backend_instance(name=backend_name)

            top_url = '{}/'.format(
                getattr(backend_instance, '_url_namespace', backend_instance.__class__.__name__)
            )

            urlpatterns += (
                url(
                    regex=r'^{}'.format(top_url), view=include(
                        backend_instance.get_url_patterns()
                    )
                ),
            )

    def __init__(self, **kwargs):
        self.kwargs = kwargs


class Variable:
    def __init__(self, name, default, environment_name):
        self.name = name
        self.default = default
        self.environment_name = environment_name

    def _get_value(self):
        return os.environ.get(self.environment_name, self.default)

    def get_value(self):
        return mark_safe(self._get_value())


class YAMLVariable(Variable):
    def _get_value(self):
        value = os.environ.get(self.environment_name)
        if value:
            value = yaml_load(stream=value)
        else:
            value = self.default

        return yaml_dump(
            data=value, allow_unicode=True, default_flow_style=True, width=999
        ).replace('...\n', '').replace('\n', '')


class PlatformTemplate:
    _registry = {}
    context = {}
    context_defaults = {}
    label = None
    name = None
    settings = None
    template_name = None
    template_string = None
    variables = None

    @classmethod
    def all(cls):
        return cls._registry.values()

    @classmethod
    def get(cls, name):
        return cls._registry[name]

    @classmethod
    def register(cls, klass):
        cls._registry[klass.name] = klass

    def __str__(self):
        return force_text(s=self.get_label())

    def get_context(self):
        return self.context

    def get_context_defaults(self):
        return self.context_defaults

    def get_label(self):
        return self.label or self.name

    def get_settings_context(self):
        result = {}
        for setting in self.settings or ():
            if setting.value:
                result[setting.global_name] = setting.value

        return result

    def get_template_name(self):
        return self.template_name or 'platform/{}.tmpl'.format(self.name)

    def get_variables_context(self):
        result = {}
        for variable in self.variables or ():
            result[variable.name] = variable.get_value()

        return result

    def render(self, context_string=None):
        """
        context_string allows the management command to pass context to this
        method as a JSON string
        """
        context = {}

        context.update(self.get_context_defaults())
        context.update(self.get_settings_context())
        context.update(self.get_variables_context())
        # get_context goes last to server as the override
        context.update(self.get_context())

        if context_string:
            context.update(
                yaml_load(stream=context_string)
            )

        if self.template_string:
            template = Template(template_string=self.template_string)
            return template.render(context=Context(dict_=context))
        else:
            return loader.render_to_string(
                template_name=self.get_template_name(),
                context=context
            )


class PlatformTemplateDockerEntrypoint(PlatformTemplate):
    label = _('Template for entrypoint.sh file inside a Docker image.')
    name = 'docker_entrypoint'
    template_name = 'platform/docker/entrypoint.tmpl'

    def get_context(self):
        context = load_env_file()
        context.update({'workers': Worker.all()})
        return context


class PlatformTemplateDockerfile(PlatformTemplate):
    label = _('Template that generates a Dockerfile file.')
    name = 'docker_dockerfile'
    template_name = 'platform/docker/dockerfile.tmpl'

    def __init__(self):
        self.variables = (
            Variable(
                name='DOCKER_LINUX_IMAGE_VERSION',
                default=DOCKER_LINUX_IMAGE_VERSION,
                environment_name='MAYAN_DOCKER_LINUX_IMAGE_VERSION'
            ),
        )


class PlatformTemplateDockerSupervisord(PlatformTemplate):
    label = _('Template for Supervisord inside a Docker image.')
    name = 'docker_supervisord'
    template_name = 'platform/docker/supervisord.tmpl'

    def get_context(self):
        return {
            'OS_USERNAME': DEFAULT_OS_USERNAME,
            'autorestart': 'false',
            'shell_path': '/bin/sh',
            'stderr_logfile': '/dev/fd/2',
            'stderr_logfile_maxbytes': '0',
            'stdout_logfile': '/dev/fd/1',
            'stdout_logfile_maxbytes': '0',
            'workers': Worker.all()
        }


class PlatformTemplateGitLabCI(PlatformTemplate):
    label = _('Template that generates a GitLab CI config file.')
    name = 'gitlab-ci'

    def __init__(self):
        self.variables = (
            Variable(
                name='DEFAULT_DATABASE_NAME',
                default=DEFAULT_DATABASE_NAME,
                environment_name='MAYAN_DEFAULT_DATABASE_NAME'
            ),
            Variable(
                name='DEFAULT_DATABASE_PASSWORD',
                default=DEFAULT_DATABASE_PASSWORD,
                environment_name='MAYAN_DEFAULT_DATABASE_PASSWORD'
            ),
            Variable(
                name='DEFAULT_DATABASE_USER',
                default=DEFAULT_DATABASE_USER,
                environment_name='MAYAN_DEFAULT_DATABASE_USER'
            ),
            Variable(
                name='DOCKER_DIND_IMAGE_VERSION',
                default=DOCKER_DIND_IMAGE_VERSION,
                environment_name='MAYAN_DOCKER_DIND_IMAGE_VERSION'
            ),
            Variable(
                name='DOCKER_LINUX_IMAGE_VERSION',
                default=DOCKER_LINUX_IMAGE_VERSION,
                environment_name='MAYAN_DOCKER_LINUX_IMAGE_VERSION'
            ),
            Variable(
                name='DOCKER_MYSQL_IMAGE_VERSION',
                default=DOCKER_MYSQL_IMAGE_VERSION,
                environment_name='MAYAN_DOCKER_MYSQL_IMAGE_VERSION'
            ),
            Variable(
                name='DOCKER_POSTGRES_IMAGE_VERSION',
                default=DOCKER_POSTGRES_IMAGE_VERSION,
                environment_name='MAYAN_DOCKER_POSTGRES_IMAGE_VERSION'
            ),
        )


class PlatformTemplateSupervisord(PlatformTemplate):
    label = _('Template for Supervisord.')
    name = 'supervisord'

    def __init__(self):
        self.variables = (
            Variable(
                name='GUNICORN_REQUESTS_JITTER',
                default=GUNICORN_REQUESTS_JITTER,
                environment_name='MAYAN_GUNICORN_REQUESTS_JITTER'
            ),
            Variable(
                name='GUNICORN_LIMIT_REQUEST_LINE',
                default=GUNICORN_LIMIT_REQUEST_LINE,
                environment_name='MAYAN_GUNICORN_GUNICORN_LIMIT_REQUEST_LINE'
            ),
            Variable(
                name='GUNICORN_MAX_REQUESTS',
                default=GUNICORN_MAX_REQUESTS,
                environment_name='MAYAN_GUNICORN_MAX_REQUESTS'
            ),
            Variable(
                name='GUNICORN_TIMEOUT',
                default=GUNICORN_TIMEOUT,
                environment_name='MAYAN_GUNICORN_TIMEOUT'
            ),
            Variable(
                name='GUNICORN_WORKER_CLASS',
                default=GUNICORN_WORKER_CLASS,
                environment_name='MAYAN_GUNICORN_WORKER_CLASS'
            ),
            Variable(
                name='GUNICORN_WORKERS',
                default=GUNICORN_WORKERS,
                environment_name='MAYAN_GUNICORN_WORKERS'
            ),
            Variable(
                name='GUNICORN_TIMEOUT',
                default=GUNICORN_TIMEOUT,
                environment_name='MAYAN_GUNICORN_TIMEOUT'
            ),
            Variable(
                name='INSTALLATION_PATH', default=DEFAULT_DIRECTORY_INSTALLATION,
                environment_name='MAYAN_INSTALLATION_PATH'
            ),
            Variable(
                name='OS_USERNAME', default=DEFAULT_OS_USERNAME,
                environment_name='MAYAN_OS_USERNAME'
            ),
            Variable(
                name='USER_SETTINGS_FOLDER',
                default=DEFAULT_USER_SETTINGS_FOLDER,
                environment_name='MAYAN_USER_SETTINGS_FOLDER'
            ),
            YAMLVariable(
                name='MEDIA_ROOT', default=settings.MEDIA_ROOT,
                environment_name='MAYAN_MEDIA_ROOT'
            )
        )

    def get_context(self):
        *_, user_settings_folder, media_root = self.variables

        return {
            'autorestart': 'true',
            'shell_path': '/bin/sh',
            'user_settings_folder': Path(
                media_root.get_value()
            ) / user_settings_folder.get_value(),
            'workers': Worker.all(),
        }


class PlatformTemplateWorkerQueues(PlatformTemplate):
    label = _('Template showing the queues of a worker.')
    name = 'worker_queues'

    variables = (
        Variable(
            name='WORKER_NAME', default=None,
            environment_name='MAYAN_WORKER_NAME'
        ),
    )

    def get_context(self):
        worker_name = self.get_variables_context().get('WORKER_NAME')
        try:
            queues = Worker.get(name=worker_name).queues
        except KeyError:
            raise KeyError('Worker name "{}" not found.'.format(worker_name))

        return {
            'queues': queues, 'queue_names': sorted(
                map(lambda x: x.name, queues)
            )
        }


PlatformTemplate.register(klass=PlatformTemplateDockerEntrypoint)
PlatformTemplate.register(klass=PlatformTemplateDockerfile)
PlatformTemplate.register(klass=PlatformTemplateDockerSupervisord)
PlatformTemplate.register(klass=PlatformTemplateGitLabCI)
PlatformTemplate.register(klass=PlatformTemplateSupervisord)
PlatformTemplate.register(klass=PlatformTemplateWorkerQueues)



from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.documents.search import document_search
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.dynamic_search.tests.mixins import SearchTestMixin

from ..permissions import permission_cabinet_view
from ..search import cabinet_search

from .mixins import CabinetTestMixin


class CabinetSearchTestCase(
    CabinetTestMixin, SearchTestMixin, GenericDocumentViewTestCase
):
    auto_create_test_cabinet = True
    test_cabinet_add_test_document = True

    def _do_test_search(self):
        return self.search_backend.search(
            search_model=cabinet_search, query={
                'documents__label': self.test_document.label
            }, user=self._test_case_user
        )

    def test_cabinet_search_no_permission(self):
        self._clear_events()

        queryset = self._do_test_search()
        self.assertTrue(self.test_cabinet not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_search_with_access(self):
        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_view
        )

        self._clear_events()

        queryset = self._do_test_search()
        self.assertTrue(self.test_cabinet in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_cabinet_search_with_access(self):
        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_view
        )

        self.test_document.delete()

        self._clear_events()

        queryset = self._do_test_search()
        self.assertTrue(self.test_cabinet in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class DocumentCabinetSearchTestCase(
    CabinetTestMixin, SearchTestMixin, GenericDocumentViewTestCase
):
    auto_create_test_cabinet = True
    test_cabinet_add_test_document = True

    def _do_test_search(self):
        return self.search_backend.search(
            search_model=document_search, query={
                'cabinets__label': self.test_cabinet.label
            }, user=self._test_case_user
        )

    def test_document_search_no_permission(self):
        self._clear_events()

        queryset = self._do_test_search()
        self.assertTrue(self.test_document not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_search_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        self._clear_events()

        queryset = self._do_test_search()
        self.assertTrue(self.test_document in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_search_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        self.test_document.delete()

        self._clear_events()

        queryset = self._do_test_search()
        self.assertTrue(self.test_document not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.core.exceptions import ValidationError

from mayan.apps.common.serialization import yaml_dump
from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.testing.tests.base import BaseTestCase

from ..models import DocumentMetadata

from .literals import (
    TEST_DATE_INVALID, TEST_PARSER_DATE_VALID, TEST_VALID_DATE,
    TEST_VALIDATOR_PATH_DATE, TEST_VALIDATOR_PATH_REGULAR_EXPRESSION,
    TEST_VALIDATOR_REGULAR_EXPRESSION_PATTERN, TEST_VALIDATOR_VALUE_INVALID,
    TEST_VALIDATOR_VALUE_VALID
)
from .mixins import MetadataTypeTestMixin


class MetadataTypeValidationTestCase(
    DocumentTestMixin, MetadataTypeTestMixin, BaseTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()

        self._create_test_document_stub()

    def test_delete_document_metadata_with_validator(self):
        """
        GitLab issue #588 "Cannot delete metadata from document if
        validator or parser is set"
        """
        self._create_test_metadata_type(add_test_document_type=True)

        self.test_metadata_type.validation = TEST_VALIDATOR_PATH_DATE
        self.test_metadata_type.save()

        document_metadata = DocumentMetadata(
            document=self.test_document, metadata_type=self.test_metadata_type,
            value=TEST_PARSER_DATE_VALID
        )

        document_metadata.full_clean()
        document_metadata.save()

        # Must not raise an error
        document_metadata.delete()

    def test_validation_date(self):
        self._create_test_metadata_type(
            add_test_document_type=True, extra_kwargs={
                'validation': TEST_VALIDATOR_PATH_DATE
            }
        )

        document_metadata = DocumentMetadata(
            document=self.test_document, metadata_type=self.test_metadata_type,
            value=TEST_DATE_INVALID
        )

        with self.assertRaises(expected_exception=ValidationError):
            # Should return error
            document_metadata.full_clean()
            document_metadata.save()

        # Should not return error
        document_metadata.value = TEST_VALID_DATE
        document_metadata.full_clean()
        document_metadata.save()

        self.assertEqual(
            getattr(
                self.test_document.metadata_value_of,
                self.test_metadata_type.name
            ), TEST_VALID_DATE
        )

    def test_validator_regular_expression(self):
        self._create_test_metadata_type(
            add_test_document_type=True, extra_kwargs={
                'validation_arguments': yaml_dump(
                    data={
                        'pattern': TEST_VALIDATOR_REGULAR_EXPRESSION_PATTERN
                    }
                ),
                'validation': TEST_VALIDATOR_PATH_REGULAR_EXPRESSION
            }
        )

        with self.assertRaises(expected_exception=ValidationError):
            self.test_metadata_type.validate_value(
                document_type=None, value=TEST_VALIDATOR_VALUE_INVALID
            )

        self.test_metadata_type.validate_value(
            document_type=None, value=TEST_VALIDATOR_VALUE_VALID
        )



from django.utils.translation import ugettext_lazy as _

from mayan.apps.smart_settings.classes import SettingNamespace

from .literals import DEFAULT_EMAIL, DEFAULT_PASSWORD, DEFAULT_USERNAME

namespace = SettingNamespace(label=_('Auto administrator'), name='autoadmin')

setting_email = namespace.add_setting(
    default=DEFAULT_EMAIL, global_name='AUTOADMIN_EMAIL', help_text=_(
        'Sets the email of the automatically created super user account.'
    )
)
setting_password = namespace.add_setting(
    default=DEFAULT_PASSWORD, global_name='AUTOADMIN_PASSWORD', help_text=_(
        'The password of the automatically created super user account. '
        'If it is equal to None, the password is randomly generated.'
    )
)
setting_username = namespace.add_setting(
    default=DEFAULT_USERNAME, global_name='AUTOADMIN_USERNAME', help_text=_(
        'The username of the automatically created super user account.'
    )
)



from actstream.models import Action

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.permissions.tests.mixins import RoleTestMixin
from mayan.apps.user_management.tests.mixins import GroupTestMixin

from ..classes import (
    EventModelRegistry, EventTypeNamespace, EventType, ModelEventType
)
from ..models import Notification
from ..permissions import permission_events_view

from .literals import (
    TEST_EVENT_TYPE_LABEL, TEST_EVENT_TYPE_NAME,
    TEST_EVENT_TYPE_NAMESPACE_LABEL, TEST_EVENT_TYPE_NAMESPACE_NAME
)


class EventsClearViewTestMixin:
    def _request_test_events_by_verb_clear_view(self):
        return self.post(
            viewname='events:events_by_verb_clear', kwargs={
                'verb': self._test_event_type.id
            }
        )

    def _request_test_events_list_clear_view(self):
        return self.post(viewname='events:events_list_clear')

    def _request_events_for_object_clear_view(self):
        return self.post(
            viewname='events:events_for_object_clear',
            kwargs=self.view_arguments
        )


class EventsExportViewTestMixin:
    def _request_test_events_by_verb_export_view(self):
        return self.post(
            viewname='events:events_by_verb_export', kwargs={
                'verb': self._test_event_type.id
            }
        )

    def _request_test_events_list_export_view(self):
        return self.post(viewname='events:events_list_export')

    def _request_events_for_object_export_view(self):
        return self.post(
            viewname='events:events_for_object_export',
            kwargs=self.view_arguments
        )


class EventListAPIViewTestMixin:
    def _request_test_event_list_api_view(self):
        return self.get(viewname='rest_api:event-list')


class EventTestMixin:
    def setUp(self):
        super().setUp()
        self.test_events = []

    def _create_test_event(self, action_object=None, actor=None, target=None):
        self.test_event = self._test_event_type.commit(
            action_object=action_object, actor=actor or self._test_case_user,
            target=target
        )
        self.test_events.append(self.test_event)


class EventTestCaseMixin:
    def setUp(self):
        super().setUp()
        Action.objects.all().delete()

    def _clear_events(self):
        Action.objects.all().delete()

    def _get_test_events(self):
        return Action.objects.all().order_by('timestamp')


class EventTypeNamespaceAPITestMixin:
    def _request_test_event_type_list_api_view(self):
        return self.get(viewname='rest_api:event-type-list')

    def _request_test_event_namespace_list_api_view(self):
        return self.get(viewname='rest_api:event-type-namespace-list')

    def _request_test_event_type_namespace_event_type_list_api_view(self):
        return self.get(
            viewname='rest_api:event-type-namespace-event-type-list',
            kwargs={
                'name': self._test_event_type_namespace.name
            }
        )


class EventTypeTestMixin:
    def setUp(self):
        super().setUp()
        self._test_event_types = []

    def _create_test_event_type(self):
        total_test_event_types = len(self._test_event_types)
        test_namespace_label = '{}_{}'.format(
            TEST_EVENT_TYPE_NAMESPACE_LABEL, total_test_event_types
        )
        test_namespace_name = '{}_{}'.format(
            TEST_EVENT_TYPE_NAMESPACE_NAME, total_test_event_types
        )
        test_event_label = '{}_{}'.format(
            TEST_EVENT_TYPE_LABEL, total_test_event_types
        )
        test_event_name = '{}_{}'.format(
            TEST_EVENT_TYPE_NAME, total_test_event_types
        )

        self._test_event_type_namespace = EventTypeNamespace(
            label=test_namespace_label,
            name=test_namespace_name
        )
        self._test_event_type = self._test_event_type_namespace.add_event_type(
            label=test_event_label,
            name=test_event_name
        )
        self._test_event_types.append(self._test_event_type)


class EventViewTestMixin:
    def _request_test_events_by_verb_view(self):
        return self.get(
            viewname='events:events_by_verb', kwargs={
                'verb': self._test_event_type.id
            }
        )

    def _request_test_events_list_view(self):
        return self.get(viewname='events:events_list')

    def _request_events_for_object_view(self):
        return self.get(
            viewname='events:events_for_object', kwargs=self.view_arguments
        )


class NotificationTestMixin(
    EventTypeTestMixin, GroupTestMixin, RoleTestMixin
):
    def _create_local_test_object(self):
        super()._create_test_object()

        EventModelRegistry.register(model=self.TestModel)

        ModelEventType.register(
            event_types=(self._test_event_type,), model=self.TestModel
        )

        EventType.refresh()

        ModelPermission.register(
            model=self.TestModel, permissions=(
                permission_events_view,
            )
        )

    def _create_local_test_user(self):
        self._create_test_user()
        self._create_test_group(add_users=(self.test_user,))
        self._create_test_role(add_groups=(self.test_group,))

    def _create_test_notification(self):
        self.test_notification = Notification.objects.create(
            user=self._test_case_user, action=Action.objects.first(),
            read=False
        )


class NotificationViewTestMixin:
    def _request_test_notification_list_view(self):
        return self.get(viewname='events:user_notifications_list')

    def _request_test_notification_mark_read_all_view(self):
        return self.post(viewname='events:notification_mark_read_all')

    def _request_test_notification_mark_read(self):
        return self.post(
            viewname='events:notification_mark_read', kwargs={
                'notification_id': self.test_notification.pk
            }
        )


class ObjectEventAPITestMixin:
    def _request_object_event_list_api_view(self):
        return self.get(
            viewname='rest_api:object-event-list',
            kwargs=self.view_arguments
        )


class UserEventViewsTestMixin:
    def _request_test_user_event_type_subscription_list_view(self):
        return self.get(viewname='events:event_types_user_subscriptions_list')



from pathlib import Path
import shutil

from django.utils.encoding import force_text

from mayan.apps.documents.storages import storage_document_files
from mayan.apps.documents.tests.base import GenericDocumentTestCase
from mayan.apps.mime_types.tests.mixins import MIMETypeBackendMixin
from mayan.apps.testing.tests.base import BaseTestCase

from ..utils import PassthroughStorageProcessor, mkdtemp, patch_files

from .mixins import StorageProcessorTestMixin


class PatchFilesTestCase(BaseTestCase):
    test_replace_text = 'replaced_text'

    def setUp(self):
        super().setUp()
        self.temporary_directory = mkdtemp()
        self.path_temporary_directory = Path(self.temporary_directory)
        self.path_test_file = self.path_temporary_directory / 'test_file.txt'

        with self.path_test_file.open(mode='w') as file_object:
            file_object.writelines(
                [
                    'line 1\n',
                    '    line 2\n',
                    'line 3\n',
                ]
            )

    def tearDown(self):
        super().tearDown()
        shutil.rmtree(path=self.temporary_directory, ignore_errors=True)

    def _patch_test_file(self):
        replace_list = [
            {
                'filename_pattern': '*',
                'content_patterns': [
                    {
                        'search': self.test_search_text,
                        'replace': self.test_replace_text,
                    }
                ]
            }
        ]
        patch_files(
            path=self.path_temporary_directory, replace_list=replace_list
        )

        with self.path_test_file.open(mode='r') as file_object:
            self.final_text = file_object.read()

    def test_file_patching_single_line(self):
        self.test_search_text = 'line 1'

        self._patch_test_file()

        self.assertEqual(self.final_text, 'replaced_text\n    line 2\nline 3\n')

    def test_file_patching_multi_line(self):
        self.test_search_text = 'line 2\nline 3\n'

        self._patch_test_file()

        self.assertEqual(self.final_text, 'line 1\n    replaced_text')

    def test_file_patching_spaces(self):
        self.test_search_text = '    line 2'

        self._patch_test_file()

        self.assertEqual(self.final_text, 'line 1\nreplaced_text\nline 3\n')

    def test_file_patching_no_matches(self):
        self.test_search_text = 'line 4'

        self._patch_test_file()

        self.assertEqual(self.final_text, 'line 1\n    line 2\nline 3\n')


class StorageProcessorTestCase(
    MIMETypeBackendMixin, StorageProcessorTestMixin, GenericDocumentTestCase
):
    auto_upload_test_document = False

    def _execute_storage_procesor(self, reverse=None):
        storage_processor = PassthroughStorageProcessor(
            app_label='documents',
            defined_storage_name=storage_document_files.name,
            log_file=force_text(s=self.path_test_file),
            model_name='DocumentFile'
        )
        storage_processor.execute(reverse=reverse)

    def _upload_and_process(self):
        self.defined_storage.dotted_path = 'django.core.files.storage.FileSystemStorage'
        self.defined_storage.kwargs = {
            'location': self.document_storage_kwargs['location']
        }

        self._upload_test_document()

        self.defined_storage.dotted_path = 'mayan.apps.storage.backends.compressedstorage.ZipCompressedPassthroughStorage'
        self.defined_storage.kwargs = {
            'next_storage_backend': 'django.core.files.storage.FileSystemStorage',
            'next_storage_backend_arguments': {
                'location': self.document_storage_kwargs['location']
            }
        }

        self._execute_storage_procesor()

    def test_processor_forwards(self):
        self._upload_and_process()

        with open(file=self.test_document.file_latest.file.path, mode='rb') as file_object:
            self.assertEqual(
                self.mime_type_backend.get_mime_type(file_object=file_object),
                ('application/zip', 'binary')
            )

        self.assertEqual(
            self.test_document.file_latest.checksum,
            self.test_document.file_latest.checksum_update(save=False)
        )

    def test_processor_forwards_and_reverse(self):
        self._upload_and_process()

        self._execute_storage_procesor(reverse=True)

        self.defined_storage.dotted_path = 'django.core.files.storage.FileSystemStorage'
        self.defined_storage.kwargs = {
            'location': self.document_storage_kwargs['location']
        }

        with open(file=self.test_document.file_latest.file.path, mode='rb') as file_object:
            self.assertNotEqual(
                self.mime_type_backend.get_mime_type(file_object=file_object),
                ('application/zip', 'binary')
            )

        self.assertEqual(
            self.test_document.file_latest.checksum,
            self.test_document.file_latest.checksum_update(save=False)
        )



from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ('documents', '0050_auto_20190725_0451'),
    ]

    operations = [
        migrations.CreateModel(
            name='WebLink',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'label', models.CharField(
                        db_index=True, help_text='A short text describing '
                        'the weblink.', max_length=96, verbose_name='Label'
                    )
                ),
                (
                    'template', models.TextField(
                        help_text='Template that will be used to craft the '
                        'final URL of the weblink. The {{ document }} '
                        'variable is available to the template.',
                        verbose_name='Template'
                    )
                ),
                (
                    'enabled', models.BooleanField(
                        default=True, verbose_name='Enabled'
                    )
                ),
                (
                    'document_types', models.ManyToManyField(
                        related_name='web_links',
                        to='documents.DocumentType',
                        verbose_name='Document types'
                    )
                ),
            ],
            options={
                'ordering': ('label',),
                'verbose_name': 'Web link',
                'verbose_name_plural': 'Web links',
            },
        ),
        migrations.CreateModel(
            name='ResolvedWebLink',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
            },
            bases=('web_links.weblink',),
        ),
    ]



class RecentDocumentViewMixin:
    """
    View mixin that automatically add a document for the view as the
    latest recently accessed document.
    """
    recent_document_view_document_property_name = 'object'

    def dispatch(self, request, *args, **kwargs):
        result = super().dispatch(request=request, *args, **kwargs)
        getattr(
            self, self.recent_document_view_document_property_name
        ).add_as_recent_document_for_user(user=request.user)

        return result



from mayan.apps.documents.search import document_search

from ..classes import SearchBackend, SearchModel

from .backends import TestSearchBackend


class SearchTestMixin:
    def _deindex_instance(self, instance):
        self.search_backend.deindex_instance(instance=instance)

    def _index_instance(self, instance):
        self.search_backend.index_instance(instance=instance)

    def _setup_test_model_search(self):
        """
        This method allows tests to add model search configurations and
        not have to import and initialize the SearchBackend.
        """

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        TestSearchBackend._test_class = cls
        cls.search_backend = SearchBackend.get_instance()

    @classmethod
    def tearDownClass(cls):
        cls.search_backend.tear_down()
        cls.search_backend.close()
        super().tearDownClass()

    def setUp(self):
        self._existing_search_models = SearchModel._registry.copy()
        super().setUp()
        # Monkeypatch the search class so that the test behavior is only
        # enabled when called from a search test.
        self._setup_test_model_search()
        SearchBackend._enable()

    def tearDown(self):
        SearchBackend._disable()
        super().tearDown()
        SearchModel._registry = self._existing_search_models


class SearchAPIViewTestMixin(SearchTestMixin):
    def _request_search_view(self):
        query = {'q': self.test_document.label}
        return self.get(
            viewname='rest_api:search-view', kwargs={
                'search_model_name': document_search.get_full_name()
            }, query=query
        )

    def _request_advanced_search_view(self):
        query = {'document_type__label': self.test_document.document_type.label}

        return self.get(
            viewname='rest_api:advanced-search-view', kwargs={
                'search_model_name': document_search.get_full_name()
            }, query=query
        )


class SearchToolsViewTestMixin(SearchTestMixin):
    def _request_search_backend_reindex_view(self):
        return self.post(viewname='search:search_backend_reindex')


class SearchViewTestMixin(SearchTestMixin):
    def _request_search_results_view(self, data, kwargs=None, query=None):
        return self.get(
            viewname='search:results', kwargs=kwargs, data=data, query=query
        )



from django.db import models, migrations


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '__first__'),
    ]

    operations = [
        migrations.CreateModel(
            name='DocumentVersionOCRError',
            fields=[
                (
                    'id', models.AutoField(
                        verbose_name='ID', serialize=False, auto_created=True,
                        primary_key=True
                    )
                ),
                (
                    'datetime_submitted', models.DateTimeField(
                        auto_now=True, verbose_name='Date time submitted',
                        db_index=True
                    )
                ),
                (
                    'result', models.TextField(
                        null=True, verbose_name='Result', blank=True
                    )
                ),
                (
                    'document_version', models.ForeignKey(
                        on_delete=models.CASCADE,
                        to='documents.DocumentVersion',
                        verbose_name='Document version'
                    )
                ),
            ],
            options={
                'ordering': ('datetime_submitted',),
                'verbose_name': 'Document Version OCR Error',
                'verbose_name_plural': 'Document Version OCR Errors',
            },
            bases=(models.Model,),
        ),
    ]



import logging

from django.apps import apps
from django.db.models.aggregates import Max, Min

from mayan.apps.lock_manager.exceptions import LockError
from mayan.celery import app

from .classes import SearchBackend, SearchModel
from .exceptions import DynamicSearchException, DynamicSearchRetry
from .literals import (
    TASK_DEINDEX_INSTANCE_MAX_RETRIES, TASK_INDEX_INSTANCE_MAX_RETRIES
)
from .settings import setting_indexing_chunk_size

logger = logging.getLogger(name=__name__)


@app.task(
    bind=True, ignore_result=True,
    max_retries=TASK_DEINDEX_INSTANCE_MAX_RETRIES, retry_backoff=True
)
def task_deindex_instance(self, app_label, model_name, object_id):
    logger.info('Executing')

    Model = apps.get_model(app_label=app_label, model_name=model_name)
    instance = Model._meta.default_manager.get(pk=object_id)

    try:
        SearchBackend.get_instance().deindex_instance(instance=instance)
    except (DynamicSearchRetry, LockError) as exception:
        raise self.retry(exc=exception)

    logger.info('Finished')


@app.task(
    bind=True, ignore_result=True, max_retries=None, retry_backoff=True
)
def task_index_search_model(self, search_model_full_name, range_string=None):
    search_model = SearchModel.get(name=search_model_full_name)

    kwargs = {
        'range_string': range_string, 'search_model': search_model
    }

    try:
        SearchBackend.get_instance().index_search_model(**kwargs)
    except Exception as exception:
        raise DynamicSearchException(
            'Unexpected error calling `index_search_model` with '
            'keyword arguments {}.'.format(kwargs)
        ) from exception


@app.task(
    bind=True, ignore_result=True,
    max_retries=TASK_INDEX_INSTANCE_MAX_RETRIES, retry_backoff=True
)
def task_index_instance(
    self, app_label, model_name, object_id, exclude_app_label=None,
    exclude_model_name=None, exclude_kwargs=None
):
    logger.info('Executing')

    Model = apps.get_model(app_label=app_label, model_name=model_name)
    if exclude_app_label and exclude_model_name:
        ExcludeModel = apps.get_model(
            app_label=exclude_app_label, model_name=exclude_model_name
        )
    else:
        ExcludeModel = None

    instance = Model._meta.default_manager.get(pk=object_id)

    try:
        SearchBackend.get_instance().index_instance(
            instance=instance, exclude_model=ExcludeModel,
            exclude_kwargs=exclude_kwargs
        )
    except (DynamicSearchRetry, LockError) as exception:
        raise self.retry(exc=exception)

    logger.info('Finished')


@app.task(ignore_result=True)
def task_reindex_backend():
    backend = SearchBackend.get_instance()
    backend.reset()

    for search_model in SearchModel.all():
        queryset = search_model.model._meta.managers_map[search_model.manager_name].all()
        if queryset:
            queryset = queryset.aggregate(min_id=Min('id'), max_id=Max('id'))
            step = setting_indexing_chunk_size.value - 1

            for id_start in range(queryset['min_id'], queryset['max_id'] + 1, step):
                range_string = '{}-{}'.format(id_start, id_start + step - 1)
                task_index_search_model.apply_async(
                    kwargs={
                        'range_string': range_string,
                        'search_model_full_name': search_model.get_full_name(),
                    }
                )



from rest_framework import status

from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.documents.tests.base import DocumentTestMixin
from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..permissions import permission_resolved_smart_link_view

from .mixins import ResolvedSmartLinkAPIViewTestMixin, SmartLinkTestMixin


class ResolvedSmartLinkAPIViewTestCase(
    DocumentTestMixin, SmartLinkTestMixin,
    ResolvedSmartLinkAPIViewTestMixin, BaseAPITestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_smart_link(add_test_document_type=True)
        self._create_test_smart_link_condition()
        self._create_test_document_stub()
        self._create_test_document_stub(label='linked')

    def test_resolved_smart_link_detail_api_view_no_permission(self):
        self._clear_events()

        response = self._request_resolved_smart_link_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_resolved_smart_link_detail_api_view_with_document_access(self):
        self.grant_access(
            obj=self.test_documents[0],
            permission=permission_resolved_smart_link_view
        )

        self._clear_events()

        response = self._request_resolved_smart_link_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_resolved_smart_link_detail_api_view_with_smart_link_access(self):
        self.grant_access(
            obj=self.test_smart_link,
            permission=permission_resolved_smart_link_view
        )

        self._clear_events()

        response = self._request_resolved_smart_link_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_resolved_smart_link_detail_api_view_with_full_access(self):
        self.grant_access(
            obj=self.test_documents[0],
            permission=permission_resolved_smart_link_view
        )
        self.grant_access(
            obj=self.test_smart_link,
            permission=permission_resolved_smart_link_view
        )

        self._clear_events()

        response = self._request_resolved_smart_link_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['label'], str(self.test_documents[0].uuid)
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_resolved_smart_link_detail_api_view_with_full_access(self):
        self.grant_access(
            obj=self.test_documents[0],
            permission=permission_resolved_smart_link_view
        )
        self.grant_access(
            obj=self.test_smart_link,
            permission=permission_resolved_smart_link_view
        )

        self.test_documents[0].delete()

        self._clear_events()

        response = self._request_resolved_smart_link_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_resolved_smart_link_list_api_view_no_permission(self):
        self._clear_events()

        response = self._request_resolved_smart_link_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_resolved_smart_link_list_api_view_with_document_access(self):
        self.grant_access(
            obj=self.test_documents[0],
            permission=permission_resolved_smart_link_view
        )

        self._clear_events()

        response = self._request_resolved_smart_link_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.data['count'], 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_resolved_smart_link_list_api_view_with_smart_link_access(self):
        self.grant_access(
            obj=self.test_smart_link,
            permission=permission_resolved_smart_link_view
        )

        self._clear_events()

        response = self._request_resolved_smart_link_list_api_view()

        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_resolved_smart_link_list_api_view_with_full_access(self):
        self.grant_access(
            obj=self.test_documents[0],
            permission=permission_resolved_smart_link_view
        )
        self.grant_access(
            obj=self.test_smart_link,
            permission=permission_resolved_smart_link_view
        )

        self._clear_events()

        response = self._request_resolved_smart_link_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 1)
        self.assertEqual(
            response.data['results'][0]['label'],
            str(self.test_documents[0].uuid)
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_resolved_smart_link_list_api_view_with_access(self):
        self.grant_access(
            obj=self.test_documents[0],
            permission=permission_resolved_smart_link_view
        )
        self.grant_access(
            obj=self.test_smart_link,
            permission=permission_resolved_smart_link_view
        )

        self.test_documents[0].delete()

        self._clear_events()

        response = self._request_resolved_smart_link_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_resolved_smart_link_document_list_api_view_no_permission(self):
        self._clear_events()

        response = self._request_resolved_smart_link_document_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_resolved_smart_link_document_list_api_view_with_main_document_access(self):
        self.grant_access(
            obj=self.test_documents[0],
            permission=permission_resolved_smart_link_view
        )

        self._clear_events()

        response = self._request_resolved_smart_link_document_list_api_view()

        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_resolved_smart_link_document_list_api_view_with_smart_link_access(self):
        self.grant_access(
            obj=self.test_smart_link,
            permission=permission_resolved_smart_link_view
        )

        self._clear_events()

        response = self._request_resolved_smart_link_document_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_resolved_smart_link_document_list_api_view_with_main_document_and_smart_link_access(self):
        self.grant_access(
            obj=self.test_documents[0],
            permission=permission_resolved_smart_link_view
        )
        self.grant_access(
            obj=self.test_smart_link,
            permission=permission_resolved_smart_link_view
        )

        self._clear_events()

        response = self._request_resolved_smart_link_document_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_resolved_smart_link_document_list_api_view_with_main_document_and_smart_link_access(self):
        self.grant_access(
            obj=self.test_documents[0],
            permission=permission_resolved_smart_link_view
        )
        self.grant_access(
            obj=self.test_smart_link,
            permission=permission_resolved_smart_link_view
        )

        self.test_documents[0].delete()

        self._clear_events()

        response = self._request_resolved_smart_link_document_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_resolved_smart_link_document_list_api_view_with_full_access(self):
        self.grant_access(
            obj=self.test_documents[0],
            permission=permission_resolved_smart_link_view
        )
        self.grant_access(
            obj=self.test_documents[1], permission=permission_document_view
        )
        self.grant_access(
            obj=self.test_smart_link,
            permission=permission_resolved_smart_link_view
        )

        self._clear_events()

        response = self._request_resolved_smart_link_document_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 1)
        self.assertEqual(
            response.data['results'][0]['label'],
            self.test_documents[1].label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_linked_document_resolved_smart_link_document_list_api_view_with_full_access(self):
        self.grant_access(
            obj=self.test_documents[0],
            permission=permission_resolved_smart_link_view
        )
        self.grant_access(
            obj=self.test_documents[1], permission=permission_document_view
        )
        self.grant_access(
            obj=self.test_smart_link,
            permission=permission_resolved_smart_link_view
        )

        self.test_documents[1].delete()

        self._clear_events()

        response = self._request_resolved_smart_link_document_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from colorful.fields import RGBColorField
import elasticsearch_dsl
import whoosh
from whoosh import qparser  # NOQA Used to initialize the whoosh.fields module.

from django.db import models

DEFAULT_ELASTICSEARCH_CLIENT_MAXSIZE = 10
DEFAULT_ELASTICSEARCH_CLIENT_SNIFF_ON_START = True
DEFAULT_ELASTICSEARCH_CLIENT_SNIFF_ON_CONNECTION_FAIL = True
DEFAULT_ELASTICSEARCH_CLIENT_SNIFFER_TIMEOUT = 60
DEFAULT_ELASTICSEARCH_HOST = 'http://127.0.0.1:9200'
DEFAULT_ELASTICSEARCH_INDICES_NAMESPACE = 'mayan'

QUERY_OPERATION_AND = 1
QUERY_OPERATION_OR = 2
TERM_OPERATION_AND = 'AND'
TERM_OPERATION_OR = 'OR'
TERM_OPERATIONS = [TERM_OPERATION_AND, TERM_OPERATION_OR]
TERM_QUOTES = ['"', '\'']
TERM_NEGATION_CHARACTER = '-'
TERM_SPACE_CHARACTER = ' '

TEXT_LOCK_INSTANCE_DEINDEX = 'dynamic_search_deindex_instance'
TEXT_LOCK_INSTANCE_INDEX = 'dynamic_search_index_instance'

# Elastic search specific.
DJANGO_TO_ELASTICSEARCH_FIELD_MAP = {
    models.AutoField: {'field': elasticsearch_dsl.field.Keyword},
    models.BooleanField: {'field': elasticsearch_dsl.field.Keyword},
    models.CharField: {'field': elasticsearch_dsl.field.Text},
    models.EmailField: {'field': elasticsearch_dsl.field.Keyword},
    models.DateTimeField: {
        'field': elasticsearch_dsl.field.Keyword,
        'transformation': lambda value: value.isoformat()
    },
    models.TextField: {'field': elasticsearch_dsl.field.Text},
    models.UUIDField: {
        'field': elasticsearch_dsl.field.Keyword, 'transformation': str
    },
    RGBColorField: {'field': elasticsearch_dsl.field.Keyword},
}

# Whoosh specific.
DJANGO_TO_WHOOSH_FIELD_MAP = {
    models.AutoField: {
        'field': whoosh.fields.ID(stored=True, unique=True),
        'transformation': str
    },
    models.CharField: {'field': whoosh.fields.TEXT},
    models.DateTimeField: {
        'field': whoosh.fields.TEXT,
        'transformation': lambda value: value.isoformat()
    },
    models.EmailField: {'field': whoosh.fields.TEXT},
    models.TextField: {'field': whoosh.fields.TEXT},
    models.UUIDField: {'field': whoosh.fields.TEXT, 'transformation': str},
    RGBColorField: {'field': whoosh.fields.TEXT},
}

WHOOSH_INDEX_DIRECTORY_NAME = 'whoosh'



from django.utils.translation import ugettext_lazy as _

from mayan.apps.smart_settings.classes import SettingNamespace

from .literals import (
    DEFAULT_MIRRORING_DOCUMENT_CACHE_LOOKUP_TIMEOUT,
    DEFAULT_MIRRORING_NODE_CACHE_LOOKUP_TIMEOUT
)

namespace = SettingNamespace(label=_('Mirroring'), name='mirroring')

setting_document_lookup_cache_timeout = namespace.add_setting(
    default=DEFAULT_MIRRORING_DOCUMENT_CACHE_LOOKUP_TIMEOUT,
    global_name='MIRRORING_DOCUMENT_CACHE_LOOKUP_TIMEOUT',
    help_text=_('Time in seconds to cache the path lookup to a document.')
)
setting_node_lookup_cache_timeout = namespace.add_setting(
    default=DEFAULT_MIRRORING_NODE_CACHE_LOOKUP_TIMEOUT,
    global_name='MIRRORING_NODE_CACHE_LOOKUP_TIMEOUT',
    help_text=_('Time in seconds to cache the path lookup to an index node.')
)



from .base import *  # NOQA

if 'silk' not in INSTALLED_APPS:  # NOQA: F405
    INSTALLED_APPS += (  # NOQA: F405
        'silk',
    )

MIDDLEWARE = (
    'silk.middleware.SilkyMiddleware',
) + MIDDLEWARE  # NOQA: F405



from django.conf.urls import url

from .api_views.workflow_instance_api_views import (
    APIWorkflowInstanceListView, APIWorkflowInstanceDetailView,
    APIWorkflowInstanceLaunchActionView,
    APIWorkflowInstanceLogEntryDetailView,
    APIWorkflowInstanceLogEntryListView,
    APIWorkflowInstanceLogEntryTransitionListView,
)
from .api_views.workflow_template_api_views import (
    APIWorkflowTemplateDetailView, APIWorkflowTemplateDocumentTypeAddView,
    APIWorkflowTemplateDocumentTypeListView,
    APIWorkflowTemplateDocumentTypeRemoveView, APIWorkflowTemplateImageView,
    APIWorkflowTemplateListView, APIWorkflowTemplateStateActionDetailView,
    APIWorkflowTemplateStateActionListView, APIWorkflowTemplateStateListView,
    APIWorkflowTemplateStateView, APIWorkflowTemplateTransitionListView,
    APIWorkflowTemplateTransitionView,
    APIWorkflowTemplateTransitionFieldDetailView,
    APIWorkflowTemplateTransitionFieldListView,
    APIWorkflowTemplateTransitionTriggerDetailView,
    APIWorkflowTemplateTransitionTriggerListView
)
from .views.workflow_instance_views import (
    WorkflowInstanceDetailView, WorkflowInstanceListView,
    WorkflowInstanceTransitionSelectView,
    WorkflowInstanceTransitionExecuteView
)
from .views.workflow_proxy_views import (
    WorkflowRuntimeProxyDocumentListView,
    WorkflowRuntimeProxyListView, WorkflowRuntimeProxyStateDocumentListView,
    WorkflowRuntimeProxyStateListView
)
from .views.workflow_template_views import (
    DocumentTypeWorkflowTemplatesView, DocumentWorkflowTemplatesLaunchView,
    ToolLaunchWorkflows, WorkflowTemplateCreateView,
    WorkflowTemplateDeleteView, WorkflowTemplateEditView,
    WorkflowTemplateLaunchView, WorkflowTemplateListView,
    WorkflowTemplatePreviewView, WorkflowTemplateDocumentTypesView
)
from .views.workflow_template_state_views import (
    WorkflowTemplateStateActionCreateView,
    WorkflowTemplateStateActionDeleteView, WorkflowTemplateStateActionEditView,
    WorkflowTemplateStateActionListView,
    WorkflowTemplateStateActionSelectionView, WorkflowTemplateStateCreateView,
    WorkflowTemplateStateDeleteView, WorkflowTemplateStateEditView,
    WorkflowTemplateStateListView
)
from .views.workflow_template_transition_views import (
    WorkflowTemplateTransitionCreateView, WorkflowTemplateTransitionDeleteView,
    WorkflowTemplateTransitionEditView, WorkflowTemplateTransitionListView,
    WorkflowTemplateTransitionTriggerEventListView,
    WorkflowTemplateTransitionFieldCreateView,
    WorkflowTemplateTransitionFieldDeleteView,
    WorkflowTemplateTransitionFieldEditView,
    WorkflowTemplateTransitionFieldListView
)

urlpatterns_workflow_instances = [
    url(
        regex=r'^documents/(?P<document_id>\d+)/workflows/$',
        name='workflow_instance_list',
        view=WorkflowInstanceListView.as_view()
    ),
    url(
        regex=r'^documents/workflows/(?P<workflow_instance_id>\d+)/$',
        name='workflow_instance_detail',
        view=WorkflowInstanceDetailView.as_view()
    ),
    url(
        regex=r'^documents/workflows/(?P<workflow_instance_id>\d+)/transitions/select/$',
        name='workflow_instance_transition_selection',
        view=WorkflowInstanceTransitionSelectView.as_view()
    ),
    url(
        regex=r'^documents/workflows/(?P<workflow_instance_id>\d+)/transitions/(?P<workflow_template_transition_id>\d+)/execute/$',
        name='workflow_instance_transition_execute',
        view=WorkflowInstanceTransitionExecuteView.as_view()
    )
]

urlpatterns_workflow_runtime_proxies = [
    url(
        regex=r'workflow_runtime_proxies/$',
        name='workflow_runtime_proxy_list',
        view=WorkflowRuntimeProxyListView.as_view()
    ),
    url(
        regex=r'^workflow_runtime_proxies/(?P<workflow_runtime_proxy_id>\d+)/documents/$',
        name='workflow_runtime_proxy_document_list',
        view=WorkflowRuntimeProxyDocumentListView.as_view()
    ),
    url(
        regex=r'^workflow_runtime_proxies/(?P<workflow_runtime_proxy_id>\d+)/states/$',
        name='workflow_runtime_proxy_state_list',
        view=WorkflowRuntimeProxyStateListView.as_view()
    ),
    url(
        regex=r'^workflow_runtime_proxies/states/(?P<workflow_runtime_proxy_state_id>\d+)/documents/$',
        name='workflow_runtime_proxy_state_document_list',
        view=WorkflowRuntimeProxyStateDocumentListView.as_view()
    )
]

urlpatterns_workflow_states = [
    url(
        regex=r'^workflow_templates/(?P<workflow_template_id>\d+)/states/$',
        name='workflow_template_state_list',
        view=WorkflowTemplateStateListView.as_view()
    ),
    url(
        regex=r'^workflow_templates/(?P<workflow_template_id>\d+)/states/create/$',
        name='workflow_template_state_create',
        view=WorkflowTemplateStateCreateView.as_view()
    ),
    url(
        regex=r'^workflow_templates/states/(?P<workflow_template_state_id>\d+)/delete/$',
        name='workflow_template_state_delete',
        view=WorkflowTemplateStateDeleteView.as_view()
    ),
    url(
        regex=r'^workflow_templates/states/(?P<workflow_template_state_id>\d+)/edit/$',
        name='workflow_template_state_edit',
        view=WorkflowTemplateStateEditView.as_view()
    )
]

urlpatterns_workflow_state_actions = [
    url(
        regex=r'^workflow_templates/states/(?P<workflow_template_state_id>\d+)/actions/$',
        name='workflow_template_state_action_list',
        view=WorkflowTemplateStateActionListView.as_view()
    ),
    url(
        regex=r'^workflow_templates/states/(?P<workflow_template_state_id>\d+)/actions/selection/$',
        name='workflow_template_state_action_selection',
        view=WorkflowTemplateStateActionSelectionView.as_view()
    ),
    url(
        regex=r'^workflow_templates/states/(?P<workflow_template_state_id>\d+)/actions/(?P<class_path>[a-zA-Z0-9_.]+)/create/$',
        name='workflow_template_state_action_create',
        view=WorkflowTemplateStateActionCreateView.as_view()
    ),
    url(
        regex=r'^workflow_templates/states/actions/(?P<workflow_template_state_action_id>\d+)/delete/$',
        name='workflow_template_state_action_delete',
        view=WorkflowTemplateStateActionDeleteView.as_view()
    ),
    url(
        regex=r'^workflow_templates/states/actions/(?P<workflow_template_state_action_id>\d+)/edit/$',
        name='workflow_template_state_action_edit',
        view=WorkflowTemplateStateActionEditView.as_view()
    )
]

urlpatterns_workflow_templates = [
    url(
        regex=r'^document_types/(?P<document_type_id>\d+)/workflow_templates/$',
        name='document_type_workflow_templates',
        view=DocumentTypeWorkflowTemplatesView.as_view()
    ),
    url(
        regex=r'^documents/(?P<document_id>\d+)/workflow_templates/launch/$',
        name='document_single_workflow_templates_launch',
        view=DocumentWorkflowTemplatesLaunchView.as_view()
    ),
    url(
        regex=r'^documents/multiple/workflow_templates/launch/$',
        name='document_multiple_workflow_templates_launch',
        view=DocumentWorkflowTemplatesLaunchView.as_view()
    ),
    url(
        regex=r'^workflow_templates/$', name='workflow_template_list',
        view=WorkflowTemplateListView.as_view()
    ),
    url(
        regex=r'^workflow_templates/create/$',
        name='workflow_template_create',
        view=WorkflowTemplateCreateView.as_view()
    ),
    url(
        regex=r'^workflow_templates/(?P<workflow_template_id>\d+)/delete/$',
        name='workflow_template_single_delete',
        view=WorkflowTemplateDeleteView.as_view()
    ),
    url(
        regex=r'^workflow_templates/multiple/delete/$',
        name='workflow_template_multiple_delete',
        view=WorkflowTemplateDeleteView.as_view()
    ),
    url(
        regex=r'^workflow_templates/(?P<workflow_template_id>\d+)/document_types/$',
        name='workflow_template_document_types',
        view=WorkflowTemplateDocumentTypesView.as_view()
    ),
    url(
        regex=r'^workflow_templates/(?P<workflow_template_id>\d+)/edit/$',
        name='workflow_template_edit',
        view=WorkflowTemplateEditView.as_view()
    ),
    url(
        regex=r'^workflow_templates/(?P<workflow_template_id>\d+)/launch/$',
        name='workflow_template_launch',
        view=WorkflowTemplateLaunchView.as_view()
    ),
    url(
        regex=r'^workflow_templates/(?P<workflow_template_id>\d+)/preview/$',
        name='workflow_template_preview',
        view=WorkflowTemplatePreviewView.as_view()
    )
]
urlpatterns_workflow_transitions = [
    url(
        regex=r'^workflow_templates/(?P<workflow_template_id>\d+)/transitions/$',
        name='workflow_template_transition_list',
        view=WorkflowTemplateTransitionListView.as_view()
    ),
    url(
        regex=r'^workflow_templates/(?P<workflow_template_id>\d+)/transitions/create/$',
        name='workflow_template_transition_create',
        view=WorkflowTemplateTransitionCreateView.as_view()
    ),
    url(
        regex=r'^workflow_templates/transitions/(?P<workflow_template_transition_id>\d+)/delete/$',
        name='workflow_template_transition_delete',
        view=WorkflowTemplateTransitionDeleteView.as_view()
    ),
    url(
        regex=r'^workflow_templates/transitions/(?P<workflow_template_transition_id>\d+)/edit/$',
        name='workflow_template_transition_edit',
        view=WorkflowTemplateTransitionEditView.as_view()
    ),
    url(
        regex=r'^workflow_templates/transitions/(?P<workflow_template_transition_id>\d+)/events/$',
        name='workflow_template_transition_triggers',
        view=WorkflowTemplateTransitionTriggerEventListView.as_view()
    )
]

urlpatterns_workflow_transition_fields = [
    url(
        regex=r'^workflow_templates/transitions/(?P<workflow_template_transition_id>\d+)/fields/create/$',
        name='workflow_template_transition_field_create',
        view=WorkflowTemplateTransitionFieldCreateView.as_view()
    ),
    url(
        regex=r'^workflow_templates/transitions/(?P<workflow_template_transition_id>\d+)/fields/$',
        name='workflow_template_transition_field_list',
        view=WorkflowTemplateTransitionFieldListView.as_view()
    ),
    url(
        regex=r'^workflow_templates/transitions/fields/(?P<workflow_template_transition_field_id>\d+)/delete/$',
        name='workflow_template_transition_field_delete',
        view=WorkflowTemplateTransitionFieldDeleteView.as_view()
    ),
    url(
        regex=r'^workflow_templates/transitions/fields/(?P<workflow_template_transition_field_id>\d+)/edit/$',
        name='workflow_template_transition_field_edit',
        view=WorkflowTemplateTransitionFieldEditView.as_view()
    )
]

urlpatterns_tools = [
    url(
        regex=r'^tools/workflows/launch/$', name='tool_launch_workflows',
        view=ToolLaunchWorkflows.as_view()
    )
]

urlpatterns = []
urlpatterns.extend(urlpatterns_tools)
urlpatterns.extend(urlpatterns_workflow_instances)
urlpatterns.extend(urlpatterns_workflow_runtime_proxies)
urlpatterns.extend(urlpatterns_workflow_states)
urlpatterns.extend(urlpatterns_workflow_state_actions)
urlpatterns.extend(urlpatterns_workflow_templates)
urlpatterns.extend(urlpatterns_workflow_transitions)
urlpatterns.extend(urlpatterns_workflow_transition_fields)

api_urls = [
    url(
        regex=r'^workflow_templates/$', name='workflow-template-list',
        view=APIWorkflowTemplateListView.as_view()
    ),
    url(
        regex=r'^workflow_templates/(?P<workflow_template_id>[0-9]+)/$',
        name='workflow-template-detail', view=APIWorkflowTemplateDetailView.as_view()
    ),
    url(
        regex=r'^workflow_templates/(?P<workflow_template_id>[0-9]+)/document_types/$',
        name='workflow-template-document-type-list',
        view=APIWorkflowTemplateDocumentTypeListView.as_view()
    ),
    url(
        regex=r'^workflow_templates/(?P<workflow_template_id>[0-9]+)/document_types/add/$',
        name='workflow-template-document-type-add',
        view=APIWorkflowTemplateDocumentTypeAddView.as_view()
    ),
    url(
        regex=r'^workflow_templates/(?P<workflow_template_id>[0-9]+)/document_types/remove/$',
        name='workflow-template-document-type-remove',
        view=APIWorkflowTemplateDocumentTypeRemoveView.as_view()
    ),
    url(
        regex=r'^workflow_templates/(?P<workflow_template_id>\d+)/image/$',
        name='workflow-template-image',
        view=APIWorkflowTemplateImageView.as_view()
    ),
    url(
        regex=r'^workflow_templates/(?P<workflow_template_id>[0-9]+)/states/$',
        name='workflow-template-state-list',
        view=APIWorkflowTemplateStateListView.as_view()
    ),
    url(
        regex=r'^workflow_templates/(?P<workflow_template_id>[0-9]+)/states/(?P<workflow_template_state_id>[0-9]+)/$',
        name='workflow-template-state-detail',
        view=APIWorkflowTemplateStateView.as_view()
    ),
    url(
        regex=r'^workflow_templates/(?P<workflow_template_id>[0-9]+)/states/(?P<workflow_template_state_id>[0-9]+)/actions/$',
        name='workflow-template-state-action-list',
        view=APIWorkflowTemplateStateActionListView.as_view()
    ),
    url(
        regex=r'^workflow_templates/(?P<workflow_template_id>[0-9]+)/states/(?P<workflow_template_state_id>[0-9]+)/actions/(?P<workflow_template_state_action_id>[0-9]+)/$',
        name='workflow-template-state-action-detail',
        view=APIWorkflowTemplateStateActionDetailView.as_view()
    ),
    url(
        regex=r'^workflow_templates/(?P<workflow_template_id>[0-9]+)/transitions/$',
        name='workflow-template-transition-list',
        view=APIWorkflowTemplateTransitionListView.as_view()
    ),
    url(
        regex=r'^workflow_templates/(?P<workflow_template_id>[0-9]+)/transitions/(?P<workflow_template_transition_id>[0-9]+)/$',
        name='workflow-template-transition-detail',
        view=APIWorkflowTemplateTransitionView.as_view()
    ),
    url(
        regex=r'^workflow_templates/(?P<workflow_template_id>[0-9]+)/transitions/(?P<workflow_template_transition_id>[0-9]+)/fields/$',
        name='workflow-template-transition-field-list',
        view=APIWorkflowTemplateTransitionFieldListView.as_view()
    ),
    url(
        regex=r'^workflow_templates/(?P<workflow_template_id>[0-9]+)/transitions/(?P<workflow_template_transition_id>[0-9]+)/fields/(?P<workflow_template_transition_field_id>[0-9]+)$',
        name='workflow-template-transition-field-detail',
        view=APIWorkflowTemplateTransitionFieldDetailView.as_view()
    ),
    url(
        regex=r'^workflow_templates/(?P<workflow_template_id>[0-9]+)/transitions/(?P<workflow_template_transition_id>[0-9]+)/triggers/$',
        name='workflow-template-transition-trigger-list',
        view=APIWorkflowTemplateTransitionTriggerListView.as_view()
    ),
    url(
        regex=r'^workflow_templates/(?P<workflow_template_id>[0-9]+)/transitions/(?P<workflow_template_transition_id>[0-9]+)/triggers/(?P<workflow_template_transition_trigger_id>[0-9]+)$',
        name='workflow-template-transition-trigger-detail',
        view=APIWorkflowTemplateTransitionTriggerDetailView.as_view()
    ),
    url(
        regex=r'^documents/(?P<document_id>[0-9]+)/workflow_instances/$',
        name='workflow-instance-list',
        view=APIWorkflowInstanceListView.as_view()
    ),
    url(
        regex=r'^documents/(?P<document_id>[0-9]+)/workflow_instances/launch/$',
        name='workflow-instance-launch',
        view=APIWorkflowInstanceLaunchActionView.as_view()
    ),
    url(
        regex=r'^documents/(?P<document_id>[0-9]+)/workflow_instances/(?P<workflow_instance_id>[0-9]+)/$',
        name='workflow-instance-detail',
        view=APIWorkflowInstanceDetailView.as_view()
    ),
    url(
        regex=r'^documents/(?P<document_id>[0-9]+)/workflow_instances/(?P<workflow_instance_id>[0-9]+)/log_entries/$',
        name='workflow-instance-log-entry-list',
        view=APIWorkflowInstanceLogEntryListView.as_view()
    ),
    url(
        regex=r'^documents/(?P<document_id>[0-9]+)/workflow_instances/(?P<workflow_instance_id>[0-9]+)/log_entries/transitions/$',
        name='workflow-instance-log-entry-transition-list',
        view=APIWorkflowInstanceLogEntryTransitionListView.as_view()
    ),
    url(
        regex=r'^documents/(?P<document_id>[0-9]+)/workflow_instances/(?P<workflow_instance_id>[0-9]+)/log_entries/(?P<workflow_instance_log_entry_id>[0-9]+)/$',
        name='workflow-instance-log-entry-detail',
        view=APIWorkflowInstanceLogEntryDetailView.as_view()
    )
]



from io import StringIO

from django.core import management

from mayan.apps.testing.tests.base import BaseTestCase

from ..classes import PlatformTemplate

from .literals import (
    TEST_TEMPLATE_LABEL, TEST_TEMPLATE_NAME, TEST_TEMPLATE_STRING,
    TEST_TEMPLATE_STRING_RENDER, TEST_TEMPLATE_STRING_RENDER_ALT,
    TEST_TEMPLATE_VARIABLE_VALUE, TEST_TEMPLATE_VARIABLE_VALUE_ALT
)


class TestPlatformTemplate(PlatformTemplate):
    context = {'test_template_variable': TEST_TEMPLATE_VARIABLE_VALUE}
    label = TEST_TEMPLATE_LABEL
    name = TEST_TEMPLATE_NAME
    template_string = TEST_TEMPLATE_STRING


PlatformTemplate.register(klass=TestPlatformTemplate)


class PlatformTemplateManagementCommandTestCase(BaseTestCase):
    def test_platform_template_simple(self):
        output = StringIO()
        args = (TEST_TEMPLATE_NAME,)
        options = {
            'stdout': output
        }
        management.call_command('platformtemplate', *args, **options)
        self.assertEqual(output.getvalue(), TEST_TEMPLATE_STRING_RENDER)

    def test_platform_template_context(self):
        output = StringIO()
        args = (
            TEST_TEMPLATE_NAME, '--context',
            'test_template_variable: {}'.format(
                TEST_TEMPLATE_VARIABLE_VALUE_ALT
            )
        )
        options = {
            'stdout': output
        }
        management.call_command('platformtemplate', *args, **options)
        self.assertEqual(output.getvalue(), TEST_TEMPLATE_STRING_RENDER_ALT)



from ...source_backends.literals import SOURCE_UNCOMPRESS_CHOICE_NEVER
from ...source_backends.web_form_backends import SourceBackendWebForm

from .base_mixins import SourceTestMixin


class WebFormSourceTestMixin(SourceTestMixin):
    _create_source_method = '_create_test_web_form_source'

    def _create_test_web_form_source(self, extra_data=None):
        backend_data = {'uncompress': SOURCE_UNCOMPRESS_CHOICE_NEVER}

        if extra_data:
            backend_data.update(extra_data)

        self._create_test_source(
            backend_path=SourceBackendWebForm.get_class_path(),
            backend_data=backend_data
        )



class WorkflowRuntimeProxyStateViewTestMixin:
    def _request_test_workflow_runtime_proxy_state_list_view(self):
        return self.get(
            viewname='document_states:workflow_runtime_proxy_state_list',
            kwargs={
                'workflow_runtime_proxy_id': self._test_workflow_template.pk
            }
        )

    def _request_test_workflow_runtime_proxy_state_document_list_view(self):
        return self.get(
            viewname='document_states:workflow_runtime_proxy_state_document_list',
            kwargs={
                'workflow_runtime_proxy_state_id': self._test_workflow_template_states[0].pk
            }
        )


class WorkflowRuntimeProxyViewTestMixin:
    def _request_test_workflow_runtime_proxy_list_view(self):
        return self.get(
            viewname='document_states:workflow_runtime_proxy_list'
        )

    def _request_test_workflow_runtime_proxy_document_list_view(self):
        return self.get(
            viewname='document_states:workflow_runtime_proxy_document_list',
            kwargs={
                'workflow_runtime_proxy_id': self._test_workflow_template.pk
            }
        )



from django import forms
from django.utils.text import format_lazy

from mayan.apps.documents.models import Document
from mayan.apps.templating.fields import ModelTemplateField

from .models import WebLink


class WebLinkForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['template'] = ModelTemplateField(
            initial_help_text=format_lazy(
                '{} ', self.fields['template'].help_text
            ), label=self.fields['template'].label, model=Document,
            model_variable='document', required=True
        )

    class Meta:
        fields = ('label', 'template', 'enabled')
        model = WebLink



from django.template.loader import get_template


class IconDriver:
    context = {}
    _registry = {}

    @classmethod
    def get(cls, name):
        return cls._registry[name]

    @classmethod
    def register(cls, driver_class):
        cls._registry[driver_class.name] = driver_class

    def get_context(self):
        return self.context

    def render(self, extra_context=None):
        context = self.get_context()
        if extra_context:
            context.update(extra_context)
        return get_template(template_name=self.template_name).render(
            context=context
        )


class FontAwesomeDriver(IconDriver):
    name = 'fontawesome'
    template_name = 'appearance/icons/font_awesome_symbol.html'

    def __init__(self, symbol):
        self.symbol = symbol

    def get_context(self):
        return {'symbol': self.symbol}


class FontAwesomeDualClassesDriver(IconDriver):
    name = 'fontawesome-dual-classes'
    template_name = 'appearance/icons/font_awesome_layers.html'

    def __init__(self, primary_class, secondary_class):
        self.primary_class = primary_class
        self.secondary_class = secondary_class

    def get_context(self):
        return {
            'data': (
                {
                    'class': 'fas fa-circle',
                    'transform': 'down-3 right-10',
                    'mask': '{}'.format(self.primary_class)
                },
                {'class': 'far fa-circle', 'transform': 'down-3 right-10'},
                {
                    'class': '{}'.format(self.secondary_class),
                    'transform': 'shrink-4 down-3 right-10'
                },
            )
        }


class FontAwesomeDualDriver(IconDriver):
    name = 'fontawesome-dual'
    template_name = 'appearance/icons/font_awesome_layers.html'

    def __init__(self, primary_symbol, secondary_symbol):
        self.primary_symbol = primary_symbol
        self.secondary_symbol = secondary_symbol

    def get_context(self):
        return {
            'data': (
                {
                    'class': 'fas fa-circle',
                    'transform': 'down-3 right-10',
                    'mask': 'fas fa-{}'.format(self.primary_symbol)
                },
                {'class': 'far fa-circle', 'transform': 'down-3 right-10'},
                {
                    'class': 'fas fa-{}'.format(self.secondary_symbol),
                    'transform': 'shrink-4 down-3 right-10'
                },
            )
        }


class FontAwesomeCSSDriver(IconDriver):
    name = 'fontawesomecss'
    template_name = 'appearance/icons/font_awesome_css.html'

    def __init__(self, css_classes):
        self.css_classes = css_classes

    def get_context(self):
        return {'css_classes': self.css_classes}


class FontAwesomeMasksDriver(IconDriver):
    name = 'fontawesome-masks'
    template_name = 'appearance/icons/font_awesome_masks.html'

    def __init__(self, data):
        self.data = data

    def get_context(self):
        return {'data': self.data}


class FontAwesomeLayersDriver(IconDriver):
    name = 'fontawesome-layers'
    template_name = 'appearance/icons/font_awesome_layers.html'

    def __init__(self, data):
        self.data = data

    def get_context(self):
        return {
            'data': self.data,
        }


class FontAwesomeStackDriver(IconDriver):
    name = 'fontawesome-stack'
    template_name = 'appearance/icons/font_awesome_stack.html'

    def __init__(self, data):
        self.data = data

    def get_context(self):
        return {
            'data': self.data,
        }


class Icon:
    def __init__(self, driver_name, **kwargs):
        self.kwargs = kwargs
        self.driver = IconDriver.get(name=driver_name)(**kwargs)

    def render(self, **kwargs):
        return self.driver.render(**kwargs)


IconDriver.register(driver_class=FontAwesomeCSSDriver)
IconDriver.register(driver_class=FontAwesomeDriver)
IconDriver.register(driver_class=FontAwesomeDualClassesDriver)
IconDriver.register(driver_class=FontAwesomeDualDriver)
IconDriver.register(driver_class=FontAwesomeLayersDriver)
IconDriver.register(driver_class=FontAwesomeMasksDriver)
IconDriver.register(driver_class=FontAwesomeStackDriver)



import hashlib
import logging
import shutil

from django.apps import apps
from django.db import models, transaction
from django.urls import reverse
from django.utils.encoding import force_text
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.classes import ModelQueryFields
from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.common.signals import signal_mayan_pre_save
from mayan.apps.converter.classes import ConverterBase
from mayan.apps.converter.exceptions import (
    InvalidOfficeFormat, PageCountError
)
from mayan.apps.events.classes import EventManagerMethodAfter
from mayan.apps.events.decorators import method_event
from mayan.apps.file_caching.models import CachePartitionFile
from mayan.apps.mime_types.classes import MIMETypeBackend
from mayan.apps.storage.classes import DefinedStorageLazy

from ..events import (
    event_document_file_created, event_document_file_deleted,
    event_document_file_downloaded, event_document_file_edited
)
from ..literals import (
    STORAGE_NAME_DOCUMENT_FILE_PAGE_IMAGE_CACHE, STORAGE_NAME_DOCUMENT_FILES
)
from ..managers import DocumentFileManager, ValidDocumentFileManager
from ..settings import setting_hash_block_size
from ..signals import (
    signal_post_document_created, signal_post_document_file_upload
)

from .document_models import Document
from .mixins import HooksModelMixin

__all__ = ('DocumentFile', 'DocumentFileSearchResult')
logger = logging.getLogger(name=__name__)


def upload_to(instance, filename):
    return instance.document.document_type.get_upload_filename(
        instance=instance, filename=filename
    )


class DocumentFile(
    ExtraDataModelMixin, HooksModelMixin, models.Model
):
    """
    Model that describes a document file and its properties
    Fields:
    * mimetype - File mimetype. MIME types are a standard way to describe the
    format of a file, in this case the file format of the document.
    Some examples: "text/plain" or "image/jpeg". Mayan uses this to determine
    how to render a document's file. More information:
    http://www.freeformatter.com/mime-types-list.html
    * encoding - File Encoding. The filesystem encoding of the document's
    file: binary 7-bit, binary 8-bit, text, base64, etc.
    * checksum - A hash/checkdigit/fingerprint generated from the document's
    binary data. Only identical documents will have the same checksum. If a
    document is modified after upload it's checksum will not match, used for
    detecting file tampering among other things.
    """
    _hooks_pre_create = []
    _pre_open_hooks = []
    _pre_save_hooks = []
    _post_save_hooks = []

    document = models.ForeignKey(
        on_delete=models.CASCADE, related_name='files', to=Document,
        verbose_name=_('Document')
    )
    timestamp = models.DateTimeField(
        auto_now_add=True, db_index=True, help_text=_(
            'The server date and time when the document file was processed.'
        ), verbose_name=_('Timestamp')
    )
    comment = models.TextField(
        blank=True, default='', help_text=_(
            'An optional short text describing the document file.'
        ), verbose_name=_('Comment')
    )
    # File related fields.
    file = models.FileField(
        storage=DefinedStorageLazy(name=STORAGE_NAME_DOCUMENT_FILES),
        upload_to=upload_to, verbose_name=_('File')
    )
    filename = models.CharField(
        blank=True, max_length=255, verbose_name=_('Filename')
    )
    mimetype = models.CharField(
        blank=True, editable=False, help_text=_(
            'The document file\'s file mimetype. MIME types are a '
            'standard way to describe the format of a file, in this case '
            'the file format of the document. Some examples: "text/plain" '
            'or "image/jpeg". '
        ), max_length=255, null=True, verbose_name=_('MIME type')
    )
    encoding = models.CharField(
        blank=True, editable=False, help_text=_(
            'The document file file encoding. binary 7-bit, binary 8-bit, '
            'text, base64, etc.'
        ), max_length=64, null=True, verbose_name=_('Encoding')
    )
    checksum = models.CharField(
        blank=True, db_index=True, editable=False, help_text=(
            'A hash/checkdigit/fingerprint generated from the document\'s '
            'binary data. Only identical documents will have the same '
            'checksum.'
        ), max_length=64, null=True, verbose_name=_('Checksum')
    )

    class Meta:
        ordering = ('timestamp',)
        verbose_name = _('Document file')
        verbose_name_plural = _('Document files')

    objects = DocumentFileManager()
    valid = ValidDocumentFileManager()

    @staticmethod
    def hash_function():
        return hashlib.sha256()

    @classmethod
    def execute_pre_create_hooks(cls, kwargs=None):
        """
        Helper method to allow checking if it is possible to create
        a new document file.
        """
        cls._execute_hooks(
            hook_list=cls._hooks_pre_create, instance=None, kwargs=kwargs
        )

    @classmethod
    def register_post_save_hook(cls, func, order=None):
        cls._insert_hook_entry(
            hook_list=cls._post_save_hooks, func=func, order=order
        )

    @classmethod
    def register_pre_create_hook(cls, func, order=None):
        cls._insert_hook_entry(
            hook_list=cls._hooks_pre_create, func=func, order=order
        )

    @classmethod
    def register_pre_open_hook(cls, func, order=None):
        cls._insert_hook_entry(
            hook_list=cls._pre_open_hooks, func=func, order=order
        )

    @classmethod
    def register_pre_save_hook(cls, func, order=None):
        cls._insert_hook_entry(
            hook_list=cls._pre_save_hooks, func=func, order=order
        )

    def __str__(self):
        return self.get_label()

    @cached_property
    def cache(self):
        Cache = apps.get_model(app_label='file_caching', model_name='Cache')
        return Cache.objects.get(
            defined_storage_name=STORAGE_NAME_DOCUMENT_FILE_PAGE_IMAGE_CACHE
        )

    @cached_property
    def cache_partition(self):
        partition, created = self.cache.partitions.get_or_create(
            name='file-{}'.format(self.uuid)
        )
        return partition

    def checksum_update(self, save=True):
        """
        Open a document file's file and update the checksum field using
        the user provided checksum function.
        """
        block_size = setting_hash_block_size.value
        if block_size == 0:
            # If the setting value is 0 that means disable read limit. To disable
            # the read limit passing None won't work, we pass -1 instead as per
            # the Python documentation.
            # https://docs.python.org/2/tutorial/inputoutput.html#methods-of-file-objects
            block_size = -1

        if self.exists():
            hash_object = DocumentFile.hash_function()
            with self.open() as file_object:
                while (True):
                    data = file_object.read(block_size)
                    if not data:
                        break

                    hash_object.update(data)

            self.checksum = force_text(s=hash_object.hexdigest())
            if save:
                self.save(update_fields=('checksum',))

            return self.checksum

    @method_event(
        event_manager_class=EventManagerMethodAfter,
        event=event_document_file_deleted,
        target='document',
    )
    def delete(self, *args, **kwargs):
        for page in self.pages.all():
            page.delete()

        name = self.file.name
        self.file.close()
        self.file.storage.delete(name=name)
        self.cache_partition.delete()

        result = super().delete(*args, **kwargs)

        if self.document.files.count() == 0:
            self.document.is_stub = False
            self.document._event_ignore = True
            self.document.save(update_fields=('is_stub',))

        return result

    def execute_pre_save_hooks(self):
        """
        Helper method to allow checking if new files are possible from
        outside the model. Currently used by the document file upload link
        condition.
        """
        DocumentFile._execute_hooks(
            hook_list=DocumentFile._pre_save_hooks, instance=self
        )

    def exists(self):
        """
        Returns a boolean value that indicates if the document's file
        exists in storage. Returns True if the document's file is verified to
        be in the document storage. This is a diagnostic flag to help users
        detect if the storage has desynchronized (ie: Amazon's S3).
        """
        name = self.file.name
        self.file.close()

        return self.file.storage.exists(name=name)

    def get_absolute_url(self):
        return reverse(
            viewname='documents:document_file_preview', kwargs={
                'document_file_id': self.pk
            }
        )

    def get_api_image_url(
        self, maximum_layer_order=None, transformation_instance_list=None,
        user=None
    ):
        first_page = self.pages.first()
        if first_page:
            return first_page.get_api_image_url(
                maximum_layer_order=maximum_layer_order,
                transformation_instance_list=transformation_instance_list,
                user=user
            )

    def get_cache_partitions(self):
        result = [self.cache_partition]
        for page in self.file_pages.all():
            result.append(page.cache_partition)

        return result

    @method_event(
        event_manager_class=EventManagerMethodAfter,
        event=event_document_file_downloaded,
        target='self',
    )
    def get_download_file_object(self):
        # Thin wrapper to make sure the normal views and API views trigger
        # then download event in the same way.
        return self.open()

    def get_intermediate_file(self):
        cache_filename = 'intermediate_file'

        try:
            cache_file = self.cache_partition.get_file(filename=cache_filename)
        except CachePartitionFile.DoesNotExist:
            logger.debug('Intermediate file not found.')

            try:
                with self.open() as file_object:
                    converter = ConverterBase.get_converter_class()(
                        file_object=file_object
                    )
                    with converter.to_pdf() as pdf_file_object:
                        with self.cache_partition.create_file(filename=cache_filename) as file_object:
                            shutil.copyfileobj(
                                fsrc=pdf_file_object, fdst=file_object
                            )

                        return self.cache_partition.get_file(filename=cache_filename).open()
            except InvalidOfficeFormat:
                return self.open()
            except Exception as exception:
                logger.error(
                    'Error creating intermediate file "%s"; %s.',
                    cache_filename, exception, exc_info=True
                )
                try:
                    cache_file = self.cache_partition.get_file(
                        filename=cache_filename
                    )
                except CachePartitionFile.DoesNotExist:
                    """Non fatal, ignore."""
                else:
                    cache_file.delete()
                raise exception
        else:
            logger.debug('Intermediate file found.')
            return cache_file.open()

    def get_label(self):
        return self.filename
    get_label.short_description = _('Label')

    def mimetype_update(self, save=True):
        """
        Read a document verions's file and determine the mimetype by using
        the MIME type backend.
        """
        if self.exists():
            try:
                with self.open() as file_object:
                    mimetype_backend = MIMETypeBackend.get_backend_instance()
                    self.mimetype, self.encoding = mimetype_backend.get_mime_type(
                        file_object=file_object
                    )
            except Exception:
                self.mimetype = ''
                self.encoding = ''
            finally:
                if save:
                    self.save(update_fields=('encoding', 'mimetype'))

    def natural_key(self):
        return (self.checksum, self.document.natural_key())
    natural_key.dependencies = ['documents.Document']

    @property
    def is_in_trash(self):
        return self.document.is_in_trash

    def open(self, raw=False):
        """
        Return a file descriptor to a document file's file irrespective of
        the storage backend.
        """
        name = self.file.name
        self.file.close()
        if raw:
            return self.file.storage.open(name=name)
        else:
            file_object = self.file.storage.open(name=name)

            result = DocumentFile._execute_hooks(
                hook_list=DocumentFile._pre_open_hooks,
                instance=self, file_object=file_object
            )

            if result:
                return result['file_object']
            else:
                return file_object

    def page_count_update(self, save=True, user=None):
        try:
            with self.open() as file_object:
                converter = ConverterBase.get_converter_class()(
                    file_object=file_object, mime_type=self.mimetype
                )
                detected_pages = converter.get_page_count()
        except PageCountError:
            """Converter backend doesn't understand the format."""
        else:
            DocumentFilePage = apps.get_model(
                app_label='documents', model_name='DocumentFilePage'
            )

            self.pages.all().delete()

            for page_number in range(detected_pages):
                DocumentFilePage.objects.create(
                    document_file=self, page_number=page_number + 1
                )

            if save:
                self._event_actor = user
                self.save()

            return detected_pages

    @property
    def pages(self):
        DocumentFilePage = apps.get_model(
            app_label='documents', model_name='DocumentFilePage'
        )
        queryset = ModelQueryFields.get(model=DocumentFilePage).get_queryset()
        return queryset.filter(pk__in=self.file_pages.values('pk'))

    @property
    def pages_first(self):
        return self.pages.first()

    def save(self, *args, **kwargs):
        """
        Overloaded save method that updates the document file's checksum,
        mimetype, and page count when created.
        """
        user = kwargs.pop('_user', self.__dict__.pop('_event_actor', None))
        new_document_file = not self.pk

        if new_document_file:
            logger.info('Creating new file for document: %s', self.document)
            DocumentFile.execute_pre_create_hooks(
                kwargs={
                    'document': self.document,
                    'file_object': self.file.open(mode='rb'),
                    'user': user
                }
            )

        try:
            self.execute_pre_save_hooks()

            signal_mayan_pre_save.send(
                instance=self, sender=DocumentFile, user=user
            )

            super().save(*args, **kwargs)

            DocumentFile._execute_hooks(
                hook_list=DocumentFile._post_save_hooks,
                instance=self
            )

            if new_document_file:
                # Only do this for new documents.
                event_document_file_created.commit(
                    actor=user, target=self, action_object=self.document
                )

                with transaction.atomic():
                    self.checksum_update(save=False)
                    self.mimetype_update(save=False)
                    self._event_actor = user
                    self.save()
                    self.page_count_update(save=False)

                    logger.info(
                        'New document file "%s" created for document: %s',
                        self, self.document
                    )

                    self.document.is_stub = False
                    if not self.document.label:
                        self.document.label = force_text(s=self.file)

                    self.document._event_ignore = True
                    self.document.save(update_fields=('is_stub', 'label'))
        except Exception as exception:
            logger.error(
                'Error creating new document file for document "%s"; %s',
                self.document, exception, exc_info=True
            )
            raise
        else:
            if new_document_file:
                signal_post_document_file_upload.send(
                    sender=DocumentFile, instance=self
                )

                if tuple(self.document.files.all()) == (self,):
                    signal_post_document_created.send(
                        instance=self.document, sender=Document
                    )
            else:
                event_document_file_edited.commit(
                    actor=user, target=self, action_object=self.document
                )

    def save_to_file(self, file_object):
        """
        Save a copy of the document from the document storage backend
        to the local filesystem.
        """
        with self.open() as input_file_object:
            shutil.copyfileobj(fsrc=input_file_object, fdst=file_object)

    @property
    def size(self):
        if self.exists():
            name = self.file.name
            self.file.close()
            return self.file.storage.size(name=name)
        else:
            return None

    @property
    def uuid(self):
        # Make cache UUID a mix of document UUID, file ID.
        return '{}-{}'.format(self.document.uuid, self.pk)


class DocumentFileSearchResult(DocumentFile):
    class Meta:
        proxy = True



import logging

from django.core.exceptions import ImproperlyConfigured

from mayan.apps.common.class_mixins import AppsModuleLoaderMixin
from mayan.apps.common.utils import get_class_full_name

logger = logging.getLogger(name=__name__)


class BackendMetaclass(type):
    def __new__(mcs, name, bases, attrs):
        new_class = super().__new__(
            mcs, name, bases, attrs
        )

        # Filter class hierarchy. Select only the class with App loader
        # support. This means that is the class meant to work as the
        # app's backend class.
        base_backend_class = [
            parent for parent in new_class.mro() if '_loader_module_name' in parent.__dict__ and parent._loader_module_name
        ]

        # No more than one app backend class should be returned.
        if len(base_backend_class) > 1:
            raise ImproperlyConfigured(
                'More than one backend parent class was returned. Should be '
                'only one.'
            )

        if base_backend_class:
            base_backend_class = base_backend_class[0]

            if base_backend_class != new_class:
                # Get this new child class full name, to be used as the
                # registry key.
                new_class_full_name = get_class_full_name(klass=new_class)

                # Initialize the app backend class registry.
                base_backend_class._registry.setdefault(
                    base_backend_class, {}
                )

                # Add the new child class to the app backend class registry.
                base_backend_class._registry[
                    base_backend_class
                ][new_class_full_name] = new_class

        return new_class


class BaseBackend(AppsModuleLoaderMixin, metaclass=BackendMetaclass):
    _registry = {}

    @classmethod
    def get(cls, name):
        return cls._registry.get(cls, {})[name]

    @classmethod
    def get_all(cls):
        return cls._registry.get(cls, ())

    @classmethod
    def get_choices(cls):
        choices = [
            (key, backend.label) for key, backend in cls.get_all().items()
        ]

        choices.sort(key=lambda x: x[1])

        return choices


class ModelBaseBackend(BaseBackend):
    def __init__(self, model_instance_id, **kwargs):
        self.model_instance_id = model_instance_id
        self.kwargs = kwargs



import logging

import sh

from django.core.files import File
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.serialization import yaml_load
from mayan.apps.storage.models import SharedUploadedFile
from mayan.apps.storage.utils import NamedTemporaryFile, touch

from ..classes import SourceBackend
from ..settings import setting_backend_arguments

from .mixins import SourceBackendInteractiveMixin, SourceBaseMixin

__all__ = ('SourceBackendSANEScanner',)
logger = logging.getLogger(name=__name__)


class SourceBackendSANEScanner(
    SourceBackendInteractiveMixin, SourceBaseMixin, SourceBackend
):
    field_order = ('device_name', 'arguments')
    fields = {
        'device_name': {
            'class': 'django.forms.CharField',
            'help_text': _(
                'Device name as returned by the SANE backend.'
            ),
            'kwargs': {'max_length': 255},
            'label': _('Device name'),
            'required': True
        },
        'arguments': {
            'class': 'django.forms.CharField',
            'help_text': _(
                'YAML formatted arguments to pass to the `scanimage` '
                'command. The arguments will change depending on the '
                'device. Execute `scanimage --help --device-name=DEVICE` '
                'for the list of supported arguments.'
            ),
            'label': _('Arguments'),
            'required': False,
        },
    }
    label = _('SANE Scanner')
    widgets = {
        'arguments': {
            'class': 'django.forms.widgets.Textarea', 'kwargs': {
                'attrs': {
                    'rows': 10
                }
            }
        }
    }

    def get_shared_uploaded_files(self):
        command_scanimage = sh.Command(
            path=setting_backend_arguments.value[
                'mayan.apps.sources.source_backends.SourceBackendSaneScanner'
            ]['scanimage_path']
        )

        with NamedTemporaryFile() as file_object:
            command_scanimage = command_scanimage.bake(
                device_name=self.kwargs['device_name'],
                format='tiff'
            )
            loaded_arguments = yaml_load(
                stream=self.kwargs.get('arguments', '{}')
            ) or {}
            # The output_file argument is only supported in version 1.0.28
            # https://gitlab.com/sane-project/backends/-/releases/1.0.28
            # Using redirection make this compatible with more versions.
            loaded_arguments['_out'] = file_object.name

            try:
                command_scanimage(**loaded_arguments)
            except sh.ErrorReturnCode:
                # The shell command is deleting the temporary file on errors.
                # Recreate it so that `NamedTemporaryFile` is able to delete
                # it when the context exits.
                touch(filename=file_object.name)
                raise
            else:
                file_object.seek(0)

                return (
                    SharedUploadedFile.objects.create(
                        file=File(file=file_object), filename='scan {}'.format(
                            now()
                        )
                    ),
                )

    def get_view_context(self, context, request):
        return {
            'subtemplates_list': [
                {
                    'name': 'appearance/generic_multiform_subtemplate.html',
                    'context': {
                        'forms': context['forms'],
                        'is_multipart': True,
                        'title': _('Document properties'),
                        'submit_label': _('Scan'),
                    },
                }
            ]
        }



import uuid

import django.core.files.storage
from django.db import migrations, models
from django.utils.encoding import force_text


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0042_auto_20180403_0702'),
    ]

    operations = [
        migrations.AlterField(
            model_name='documentversion',
            name='encoding',
            field=models.CharField(
                blank=True, editable=False, max_length=64, null=True,
                verbose_name='Encoding'
            ),
        ),
        migrations.AlterField(
            model_name='documentversion',
            name='file',
            field=models.FileField(
                storage=django.core.files.storage.FileSystemStorage(
                    location=b'/home/rosarior/development/mayan-edms/mayan/media/document_storage'
                ), upload_to=force_text(s=uuid.uuid4()), verbose_name='File'
            ),
        ),
        migrations.AlterField(
            model_name='documentversion',
            name='mimetype',
            field=models.CharField(
                blank=True, editable=False, max_length=255, null=True,
                verbose_name='MIME type'
            ),
        ),
    ]



from mayan.apps.common.utils import get_class_full_name

from ..authentication_backends import (
    AuthenticationBackendModelEmailPassword,
    AuthenticationBackendModelUsernamePassword
)

PATH_AUTHENTICATION_BACKEND_EMAIL = get_class_full_name(
    klass=AuthenticationBackendModelEmailPassword
)
PATH_AUTHENTICATION_BACKEND_USERNAME = get_class_full_name(
    klass=AuthenticationBackendModelUsernamePassword
)

TEST_PASSWORD_NEW = 'new_password_123'



from django.utils.translation import ugettext_lazy as _

from mayan.apps.navigation.classes import Link

from .icons import (
    icon_document_file_attachment_send_single,
    icon_document_file_send_link_single,
    icon_document_file_attachment_send_multiple,
    icon_document_file_send_link_multiple,
    icon_document_send_link_single, icon_document_send_link_multiple,
    icon_document_version_send_attachment_single,
    icon_document_version_send_link_single,
    icon_document_version_send_link_multiple,
    icon_document_version_send_attachment_multiple, icon_user_mailer_create,
    icon_user_mailer_delete, icon_user_mailer_edit, icon_user_mailer_list,
    icon_user_mailer_setup, icon_user_mailer_test
)
from .permissions import (
    permission_send_document_file_attachment,
    permission_send_document_file_link,
    permission_send_document_version_attachment,
    permission_send_document_version_link, permission_send_document_link,
    permission_user_mailer_create, permission_user_mailer_delete,
    permission_user_mailer_edit, permission_user_mailer_use,
    permission_user_mailer_view
)

# Document

link_send_document_link_single = Link(
    args='resolved_object.pk', icon=icon_document_send_link_single,
    permissions=(permission_send_document_link,),
    text=_('Email document link'), view='mailer:send_document_link_single'
)
link_send_document_link_multiple = Link(
    icon=icon_document_send_link_multiple, text=_('Email document link'),
    view='mailer:send_document_link_multiple'
)

# Document file

link_send_document_file_attachment_single = Link(
    args='resolved_object.pk',
    icon=icon_document_file_attachment_send_single,
    permissions=(permission_send_document_file_attachment,),
    text=_('Email document file'),
    view='mailer:send_document_file_attachment_single'
)
link_send_document_file_attachment_multiple = Link(
    icon=icon_document_file_attachment_send_multiple,
    text=_('Email document file'),
    view='mailer:send_document_file_attachment_multiple'
)
link_send_document_file_link_single = Link(
    args='resolved_object.pk', icon=icon_document_file_send_link_single,
    permissions=(permission_send_document_file_link,),
    text=_('Email document file link'),
    view='mailer:send_document_file_link_single'
)
link_send_document_file_link_multiple = Link(
    icon=icon_document_file_send_link_multiple,
    text=_('Email document file link'),
    view='mailer:send_document_file_link_multiple'
)

# Document version

link_send_document_version_attachment_single = Link(
    args='resolved_object.pk',
    icon=icon_document_version_send_attachment_single,
    permissions=(permission_send_document_version_attachment,),
    text=_('Email document version'),
    view='mailer:send_document_version_attachment_single'
)
link_send_document_version_attachment_multiple = Link(
    icon=icon_document_version_send_attachment_multiple,
    text=_('Email document version'),
    view='mailer:send_document_version_attachment_multiple'
)
link_send_document_version_link_single = Link(
    args='resolved_object.pk', icon=icon_document_version_send_link_single,
    permissions=(permission_send_document_version_link,),
    text=_('Email document version link'),
    view='mailer:send_document_version_link_single'
)
link_send_document_version_link_multiple = Link(
    icon=icon_document_version_send_link_multiple,
    text=_('Email link version'),
    view='mailer:send_document_version_link_multiple'
)

# Mailing profile

link_user_mailer_create = Link(
    icon=icon_user_mailer_create,
    permissions=(permission_user_mailer_create,),
    text=_('Create mailing profile'),
    view='mailer:user_mailer_backend_selection'
)
link_user_mailer_delete = Link(
    args='resolved_object.pk', icon=icon_user_mailer_delete,
    permissions=(permission_user_mailer_delete,), tags='dangerous',
    text=_('Delete'), view='mailer:user_mailer_delete'
)
link_user_mailer_edit = Link(
    args='object.pk', icon=icon_user_mailer_edit,
    permissions=(permission_user_mailer_edit,), text=_('Edit'),
    view='mailer:user_mailer_edit'
)
link_user_mailer_list = Link(
    icon=icon_user_mailer_list,
    permissions=(permission_user_mailer_view,),
    text=_('Mailing profiles list'), view='mailer:user_mailer_list'
)
link_user_mailer_setup = Link(
    icon=icon_user_mailer_setup,
    permissions=(permission_user_mailer_view,), text=_('Mailing profiles'),
    view='mailer:user_mailer_list'
)
link_user_mailer_test = Link(
    args='object.pk', icon=icon_user_mailer_test,
    permissions=(permission_user_mailer_use,), text=_('Test'),
    view='mailer:user_mailer_test'
)



def condition_user_is_not_super_user(context, resolved_object):
    #TODO: Use a subclass test instead
    if hasattr(resolved_object, 'is_staff'):
        user = resolved_object
        return not user.is_superuser and not user.is_staff
    return True



from django.http.response import StreamingHttpResponse

from rest_framework import status

from mayan.apps.documents.events import (
    event_document_created, event_document_file_created,
    event_document_file_edited, event_document_version_created,
    event_document_version_page_created
)
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.permissions import permission_document_create
from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.rest_api.tests.base import BaseAPITestCase

from .mixins.staging_folder_source_mixins import (
    StagingFolderActionAPIViewTestMixin, StagingFolderTestMixin
)


class StagingFolderActionAPIViewTestCase(
    DocumentTestMixin, StagingFolderTestMixin,
    StagingFolderActionAPIViewTestMixin, BaseAPITestCase
):
    auto_create_test_document_type = False
    auto_upload_test_document = False

    def test_staging_folder_file_delete_action_api_view_no_permission(self):
        self._copy_test_staging_folder_document()

        test_staging_folder = self.test_source.get_backend_instance()

        staging_folder_file_count = len(
            list(test_staging_folder.get_files())
        )

        self._clear_events()

        response = self._request_test_staging_folder_file_delete_action_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            len(
                list(test_staging_folder.get_files())
            ), staging_folder_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_staging_folder_file_delete_action_api_view_with_access(self):
        self.grant_access(
            obj=self.test_source, permission=permission_document_create
        )

        self._copy_test_staging_folder_document()

        test_staging_folder = self.test_source.get_backend_instance()

        staging_folder_file_count = len(
            list(test_staging_folder.get_files())
        )

        self._clear_events()

        response = self._request_test_staging_folder_file_delete_action_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            len(
                list(test_staging_folder.get_files())
            ), staging_folder_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_staging_folder_file_image_action_api_view_no_permission(self):
        self._copy_test_staging_folder_document()

        self._clear_events()

        response = self._request_test_staging_folder_file_image_action_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_staging_folder_file_image_action_api_view_with_access(self):
        self.grant_access(
            obj=self.test_source, permission=permission_document_create
        )

        self._copy_test_staging_folder_document()

        self._clear_events()

        response = self._request_test_staging_folder_file_image_action_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertTrue(isinstance(response, StreamingHttpResponse))

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_staging_folder_file_list_action_api_view_no_permission(self):
        self._copy_test_staging_folder_document()

        self._clear_events()

        response = self._request_test_staging_folder_file_list_action_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_staging_folder_file_list_action_api_view_with_access(self):
        self.grant_access(
            obj=self.test_source, permission=permission_document_create
        )

        self._copy_test_staging_folder_document()

        self._clear_events()

        response = self._request_test_staging_folder_file_list_action_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            response.data[0]['encoded_filename'],
            self.test_staging_folder_file.encoded_filename
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_staging_folder_file_upload_api_view_no_permission(self):
        self._create_test_document_type()

        self._copy_test_staging_folder_document()

        document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_staging_folder_file_upload_action_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(Document.objects.count(), document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_staging_folder_file_upload_api_view_with_document_type_access(self):
        self._create_test_document_type()

        self.grant_access(
            obj=self.test_document_type, permission=permission_document_create
        )

        self._copy_test_staging_folder_document()

        document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_staging_folder_file_upload_action_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(Document.objects.count(), document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_staging_folder_file_upload_api_view_with_source_access(self):
        self._create_test_document_type()

        self.grant_access(
            obj=self.test_source, permission=permission_document_create
        )

        self._copy_test_staging_folder_document()

        document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_staging_folder_file_upload_action_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(Document.objects.count(), document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_staging_folder_file_upload_api_view_with_full_access(self):
        self._create_test_document_type()

        self.grant_access(
            obj=self.test_document_type, permission=permission_document_create
        )
        self.grant_access(
            obj=self.test_source, permission=permission_document_create
        )

        self._copy_test_staging_folder_document()

        document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_staging_folder_file_upload_action_api_view()
        self.assertEqual(response.status_code, status.HTTP_202_ACCEPTED)

        self.assertEqual(Document.objects.count(), document_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 5)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self.test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_version)
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )



from django.db import connection
from django.utils.translation import ugettext_lazy as _

from mayan.apps.dynamic_search.classes import SearchModel

from .permissions import permission_index_instance_view


def transformation_format_uuid(term_string):
    if connection.vendor in ('mysql', 'sqlite'):
        return term_string.replace('-', '')
    else:
        return term_string


index_instance_node_search = SearchModel(
    app_label='document_indexing', model_name='IndexInstanceNodeSearchResult',
    permission=permission_index_instance_view,
    serializer_path='mayan.apps.document_indexing.serializers.IndexInstanceNodeSerializer'
)
index_instance_node_search.add_proxy_model(
    app_label='document_indexing', model_name='IndexInstanceNode'
)

index_instance_node_search.add_model_field(
    field='value', label=_('Value')
)

index_instance_node_search.add_model_field(
    field='documents__document_type__label', label=_('Document type')
)
index_instance_node_search.add_model_field(
    field='documents__files__mimetype', label=_('Document MIME type')
)
index_instance_node_search.add_model_field(
    field='documents__label', label=_('Document label')
)
index_instance_node_search.add_model_field(
    field='documents__description', label=_('Document description')
)
index_instance_node_search.add_model_field(
    field='documents__uuid', label=_('Document UUID'),
    transformation_function=transformation_format_uuid
)
index_instance_node_search.add_model_field(
    field='documents__files__checksum', label=_('Document checksum')
)



from rest_framework import status

from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..events import event_announcement_created, event_announcement_edited
from ..models import Announcement
from ..permissions import (
    permission_announcement_create, permission_announcement_delete,
    permission_announcement_edit, permission_announcement_view
)

from .literals import TEST_ANNOUNCEMENT_LABEL, TEST_ANNOUNCEMENT_TEXT
from .mixins import AnnouncementAPIViewTestMixin, AnnouncementTestMixin


class AnnouncementAPIViewTestCase(
    AnnouncementAPIViewTestMixin, AnnouncementTestMixin, BaseAPITestCase
):
    def test_announcement_create_api_view_no_permission(self):
        self._clear_events()

        response = self._request_announcement_create_view()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(Announcement.objects.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_announcement_create_api_view_with_permission(self):
        self.grant_permission(permission=permission_announcement_create)

        self._clear_events()

        response = self._request_announcement_create_view()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        announcement = Announcement.objects.first()
        self.assertEqual(response.data['id'], announcement.pk)
        self.assertEqual(response.data['label'], TEST_ANNOUNCEMENT_LABEL)
        self.assertEqual(response.data['text'], TEST_ANNOUNCEMENT_TEXT)

        self.assertEqual(Announcement.objects.count(), 1)
        self.assertEqual(announcement.label, TEST_ANNOUNCEMENT_LABEL)
        self.assertEqual(announcement.text, TEST_ANNOUNCEMENT_TEXT)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_announcement)
        self.assertEqual(events[0].verb, event_announcement_created.id)

    def test_announcement_delete_api_view_no_permission(self):
        self._create_test_announcement()

        self._clear_events()

        response = self._request_announcement_delete_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(Announcement.objects.count(), 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_announcement_delete_api_view_with_access(self):
        self._create_test_announcement()
        self.grant_access(
            obj=self.test_announcement,
            permission=permission_announcement_delete
        )

        self._clear_events()

        response = self._request_announcement_delete_view()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(Announcement.objects.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_announcement_detail_api_view_no_permission(self):
        self._create_test_announcement()

        self._clear_events()

        response = self._request_announcement_detail_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_announcement_detail_api_view_with_access(self):
        self._create_test_announcement()
        self.grant_access(
            obj=self.test_announcement,
            permission=permission_announcement_view
        )

        self._clear_events()

        response = self._request_announcement_detail_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.data['label'], self.test_announcement.label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_announcement_edit_api_view_via_patch_view_no_permission(self):
        self._create_test_announcement()

        test_announcement_label = self.test_announcement.label
        test_announcement_text = self.test_announcement.text

        self._clear_events()

        response = self._request_announcement_edit_via_patch_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.test_announcement.refresh_from_db()
        self.assertEqual(
            self.test_announcement.label, test_announcement_label
        )
        self.assertEqual(self.test_announcement.text, test_announcement_text)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_announcement_edit_api_view_via_patch_view_with_access(self):
        self._create_test_announcement()
        self.grant_access(
            obj=self.test_announcement, permission=permission_announcement_edit
        )

        test_announcement_label = self.test_announcement.label
        test_announcement_text = self.test_announcement.text

        self._clear_events()

        response = self._request_announcement_edit_via_patch_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.test_announcement.refresh_from_db()
        self.assertNotEqual(
            self.test_announcement.label, test_announcement_label
        )
        self.assertNotEqual(self.test_announcement.text, test_announcement_text)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_announcement)
        self.assertEqual(events[0].verb, event_announcement_edited.id)

    def test_announcement_edit_api_view_via_put_view_no_permission(self):
        self._create_test_announcement()

        test_announcement_label = self.test_announcement.label
        test_announcement_text = self.test_announcement.text

        self._clear_events()

        response = self._request_announcement_edit_via_put_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.test_announcement.refresh_from_db()
        self.assertEqual(
            self.test_announcement.label, test_announcement_label
        )
        self.assertEqual(self.test_announcement.text, test_announcement_text)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_announcement_edit_api_view_via_put_view_with_access(self):
        self._create_test_announcement()
        self.grant_access(
            obj=self.test_announcement, permission=permission_announcement_edit
        )

        test_announcement_label = self.test_announcement.label
        test_announcement_text = self.test_announcement.text

        self._clear_events()

        response = self._request_announcement_edit_via_put_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.test_announcement.refresh_from_db()
        self.test_announcement.refresh_from_db()
        self.assertNotEqual(
            self.test_announcement.label, test_announcement_label
        )
        self.assertNotEqual(self.test_announcement.text, test_announcement_text)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_announcement)
        self.assertEqual(events[0].verb, event_announcement_edited.id)



from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.documents.serializers.document_serializers import DocumentSerializer
from mayan.apps.rest_api import generics
from mayan.apps.rest_api.api_view_mixins import ExternalObjectAPIViewMixin

from .models import Cabinet
from .permissions import (
    permission_cabinet_add_document, permission_cabinet_create,
    permission_cabinet_delete, permission_cabinet_edit,
    permission_cabinet_remove_document, permission_cabinet_view
)
from .serializers import (
    CabinetDocumentAddSerializer, CabinetDocumentRemoveSerializer,
    CabinetSerializer
)


class APIDocumentCabinetListView(
    ExternalObjectAPIViewMixin, generics.ListAPIView
):
    """
    Returns a list of all the cabinets to which a document belongs.
    """
    external_object_queryset = Document.valid.all()
    external_object_pk_url_kwarg = 'document_id'
    mayan_external_object_permissions = {'GET': (permission_cabinet_view,)}
    mayan_object_permissions = {'GET': (permission_cabinet_view,)}
    serializer_class = CabinetSerializer

    def get_queryset(self):
        return self.external_object.cabinets.all()


class APICabinetListView(generics.ListCreateAPIView):
    """
    get: Returns a list of all the cabinets.
    post: Create a new cabinet.
    """
    mayan_object_permissions = {'GET': (permission_cabinet_view,)}
    mayan_view_permissions = {'POST': (permission_cabinet_create,)}
    ordering_fields = ('id', 'label')
    queryset = Cabinet.objects.all()
    serializer_class = CabinetSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class APICabinetView(generics.RetrieveUpdateDestroyAPIView):
    """
    delete: Delete the selected cabinet.
    get: Returns the details of the selected cabinet.
    patch: Edit the selected cabinet.
    put: Edit the selected cabinet.
    """
    lookup_url_kwarg = 'cabinet_id'
    mayan_object_permissions = {
        'GET': (permission_cabinet_view,),
        'PUT': (permission_cabinet_edit,),
        'PATCH': (permission_cabinet_edit,),
        'DELETE': (permission_cabinet_delete,)
    }
    queryset = Cabinet.objects.all()
    serializer_class = CabinetSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class APICabinetDocumentAddView(generics.ObjectActionAPIView):
    """
    post: Add a document to a cabinet.
    """
    lookup_url_kwarg = 'cabinet_id'
    mayan_object_permissions = {
        'POST': (permission_cabinet_add_document,)
    }
    serializer_class = CabinetDocumentAddSerializer
    queryset = Cabinet.objects.all()

    def object_action(self, request, serializer):
        document = serializer.validated_data['document']
        self.object._event_actor = self.request.user
        self.object.document_add(document=document)


class APICabinetDocumentRemoveView(generics.ObjectActionAPIView):
    """
    post: Remove a document from a cabinet.
    """
    lookup_url_kwarg = 'cabinet_id'
    mayan_object_permissions = {
        'POST': (permission_cabinet_remove_document,)
    }
    serializer_class = CabinetDocumentRemoveSerializer
    queryset = Cabinet.objects.all()

    def object_action(self, request, serializer):
        document = serializer.validated_data['document']
        self.object._event_actor = self.request.user
        self.object.document_remove(document=document)


class APICabinetDocumentListView(
    ExternalObjectAPIViewMixin, generics.ListAPIView
):
    """
    get: Returns a list of all the documents contained in a particular cabinet.
    """
    external_object_class = Cabinet
    external_object_pk_url_kwarg = 'cabinet_id'
    mayan_external_object_permissions = {'GET': (permission_cabinet_view,)}
    mayan_object_permissions = {
        'GET': (permission_document_view,),
    }
    serializer_class = DocumentSerializer

    def get_queryset(self):
        return Document.valid.filter(
            pk__in=self.external_object.documents.only('pk')
        )



from django.http import Http404

from mayan.apps.acls.models import AccessControlList

from ..literals import FIELDS_ALL, FIELDS_USER


class DynamicUserViewMixin:
    def dispatch(self, request, *args, **kwargs):
        object_raw = self.get_object(queryset=self.get_source_queryset())
        self.is_current_user = object_raw == self.request.user

        queryset = AccessControlList.objects.restrict_queryset(
            permission=self.object_permission,
            queryset=self.get_source_queryset(),
            user=self.request.user
        )

        try:
            filtered_obj = self.get_object(queryset=queryset)
        except Http404:
            filtered_obj = None

        self.user_has_access = filtered_obj == self.request.user

        return super().dispatch(request=request, *args, **kwargs)

    def get_object_permission(self):
        if self.is_current_user:
            return
        else:
            return self.object_permission


class DynamicUserFormFieldViewMixin(DynamicUserViewMixin):
    def get_form_class(self):
        if self.is_current_user:
            if self.user_has_access:
                self.fields = FIELDS_ALL
            else:
                self.fields = FIELDS_USER
        else:
            self.fields = FIELDS_ALL

        return super().get_form_class()


class DynamicExternalUserViewMixin:
    def dispatch(self, request, *args, **kwargs):
        external_object_raw = self.get_external_object(
            queryset=self.get_external_object_queryset()
        )
        self.is_current_user = external_object_raw == self.request.user

        queryset = AccessControlList.objects.restrict_queryset(
            permission=self.external_object_permission,
            queryset=self.get_external_object_queryset(),
            user=self.request.user
        )

        try:
            external_object_filtered = self.get_external_object(
                queryset=queryset
            )
        except Http404:
            external_object_filtered = None

        self.user_has_access = external_object_filtered == self.request.user

        return super().dispatch(request=request, *args, **kwargs)

    def get_external_object_permission(self):
        if self.is_current_user:
            return
        else:
            return self.external_object_permission



from mayan.apps.testing.tests.base import GenericViewTestCase

from ..events import event_source_created, event_source_edited
from ..models import Source
from ..permissions import (
    permission_sources_create, permission_sources_delete,
    permission_sources_edit, permission_sources_view
)

from .literals import TEST_SOURCE_LABEL
from .mixins.base_mixins import SourceTestMixin, SourceViewTestMixin


class SourceViewTestCase(
    SourceTestMixin, SourceViewTestMixin, GenericViewTestCase
):
    auto_create_test_source = False

    def test_source_create_view_no_permission(self):
        source_count = Source.objects.count()

        self._clear_events()

        response = self._request_test_source_create_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(Source.objects.count(), source_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_create_view_with_permission(self):
        self.grant_permission(permission=permission_sources_create)

        source_count = Source.objects.count()

        self._clear_events()

        response = self._request_test_source_create_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(self.test_source.label, TEST_SOURCE_LABEL)
        self.assertEqual(Source.objects.count(), source_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_source)
        self.assertEqual(events[0].verb, event_source_created.id)

    def test_source_delete_view_no_permission(self):
        self._create_test_source()

        source_count = Source.objects.count()

        self._clear_events()

        response = self._request_test_source_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Source.objects.count(), source_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_delete_view_with_access(self):
        self._create_test_source()

        self.grant_access(
            obj=self.test_source, permission=permission_sources_delete
        )

        source_count = Source.objects.count()

        self._clear_events()

        response = self._request_test_source_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Source.objects.count(), source_count - 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_edit_view_no_permission(self):
        self._create_test_source()
        test_instance_values = self._model_instance_to_dictionary(
            instance=self.test_source
        )

        self._clear_events()

        response = self._request_test_source_edit_view()
        self.assertEqual(response.status_code, 404)

        self.test_source.refresh_from_db()
        self.assertEqual(
            self._model_instance_to_dictionary(
                instance=self.test_source
            ), test_instance_values
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_edit_view_with_access(self):
        self._create_test_source()
        test_instance_values = self._model_instance_to_dictionary(
            instance=self.test_source
        )
        self.grant_access(
            obj=self.test_source, permission=permission_sources_edit
        )

        self._clear_events()

        response = self._request_test_source_edit_view()
        self.assertEqual(response.status_code, 302)

        self.test_source.refresh_from_db()
        self.assertNotEqual(
            self._model_instance_to_dictionary(
                instance=self.test_source
            ), test_instance_values
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_source)
        self.assertEqual(events[0].verb, event_source_edited.id)

    def test_source_list_view_no_permission(self):
        self._create_test_source()

        self._clear_events()

        response = self._request_test_source_list_view()
        self.assertEqual(response.status_code, 200)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_list_view_with_access(self):
        self._create_test_source()

        self.grant_access(
            obj=self.test_source, permission=permission_sources_view
        )

        self._clear_events()

        response = self._request_test_source_list_view()
        self.assertContains(
            response=response, text=self.test_source.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_test_get_view_no_permission(self):
        self._create_test_source()

        self._clear_events()

        response = self._request_test_source_test_get_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_test_get_view_with_access(self):
        self._create_test_source()

        self.grant_access(
            obj=self.test_source, permission=permission_sources_edit
        )

        self._clear_events()

        response = self._request_test_source_test_get_view()
        self.assertEqual(response.status_code, 200)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_test_post_view_no_permission(self):
        self._create_test_source()

        self._clear_events()

        response = self._request_test_source_test_post_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_test_post_view_with_access(self):
        self._create_test_source()

        self.grant_access(
            obj=self.test_source, permission=permission_sources_edit
        )

        self._clear_events()

        response = self._request_test_source_test_post_view()
        self.assertEqual(response.status_code, 302)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('storage', '0002_auto_20200528_0826'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='StorageSharedUploadedFile',
            new_name='SharedUploadedFile',
        ),
    ]
